#!/usr/bin/python


import re
import os.path


GWT = re.compile('^\s+(Given|When|Then) (.+)\s+$')


def parse(file):
    feature = {'title': None,
               'description': None,
               'scenarios': [],
               'file': file.name}
    cursor = None
    is_comment = False
    current_scenario = {'title': None,
                        'steps': []}
    for l in file.readlines():
        if l.startswith('Feature'):
            cursor = 'Feature'
            feature['title'] = l.strip()[len('Feature: '):]
            continue
        elif l.strip().startswith('Scenario: '):
            cursor = 'Scenario'
            current_scenario['title'] = l.strip()[len('Scenario: '):]
            continue
        elif GWT.match(l):
            clause, content = GWT.match(l).groups()
            step = {'clause': clause,
                    'content': content}
            current_scenario['steps'].append(step)
            continue
        else:
            if l.strip() in ['', '\n'] and is_comment is False:
                if cursor == 'Scenario':
                    # archive scenario
                    feature['scenarios'].append(current_scenario)
                    current_scenario = {'title': None,
                                        'steps': []}
                # reset cursor
                cursor = None
            else:
                line = l.strip()
                if len(line) < 1:
                    line = '\n'
                if cursor == 'Feature':
                    feature['description'] = line
                elif cursor == 'Scenario':
                    if '"""' in l:
                        if is_comment:
                            # end of comment
                            is_comment = False
                        else:
                            is_comment = True
                            line = '\n' + line
                    current_scenario['steps'][-1]['content'] += line
                    if is_comment:
                        current_scenario['steps'][-1]['content'] += '\n'
    # close any unfinished business
    if cursor == 'Scenario':
        feature['scenarios'].append(current_scenario)
    return feature


def dict2adoc(feature, name=None):
    if not name:
        name = feature['title']
    adoc = """%s(7)
%s
:doctype: manpage


NAME
----
%s - autogenerated from feature spec `%s`


SYNOPSIS
--------

%s


""" % (feature['title'],
       '=' * len(feature['title']) + '===',
       name,
       os.path.basename(feature['file']),
       feature.get('description', '.'))
    adoc += "SCENARIOS\n---------\n\n"
    for scenario in feature.get('scenarios', []):
        adoc += "%s\n%s\n\n" % (scenario['title'],
                                '~' * len(scenario['title']))
        for step in scenario.get('steps', []):
            if '\n' in step['content']:
                fl, rest = step['content'].split('\n', 1)
            else:
                fl, rest = step['content'], None
            adoc += "* *_%s_* %s" % (step['clause'], fl)
            if rest:
                adoc += "\n....\n%s\n....\n" % rest
            else:
                adoc += ",\n"
        if adoc.endswith('....\n'):
            adoc += '\n'
        else:
            adoc = adoc[:-2] + '\n\n'
    return adoc


if __name__ == '__main__':
    import sys
    import glob

    path = sys.argv[1]
    for f in glob.glob(path):
        # foo.feature -> rdopkg.feature-foo.7.adoc
        basefn = re.sub(r'(.+)\.feature$',
                        'rdopkg-feature-\g<1>',
                        f.split('/')[-1])
        output = basefn + '.7.adoc'
        with open(f) as feature, open(output, 'w') as adoc:
            adoc.write(dict2adoc(parse(feature), name=basefn))



# -*- encoding: utf-8 -*-
from __future__ import unicode_literals

from rdopkg.cli import rdopkg
from rdopkg.utils.git import git, git_branch
from rdopkg.utils import log

import test_common as common
from test_common import DIST_POSTFIX

import pytest

RPM_AVAILABLE = False
try:
    import rpm  # NOQA
    RPM_AVAILABLE = True
except ImportError:
    pass


def _test_patch(asset, version, dir):
    dist_path = common.prep_spec_test(dir, asset)
    log.log.setLevel(log.WARN)

    with dist_path.as_cwd():
        spec_version, spec_release_parts, spec_milestone = version
        tag = spec_version
        if spec_milestone:
            tag += spec_milestone

        common.prep_patches_branch(tag=tag)
        commit_before = git('rev-parse', 'HEAD')
        common.add_patches()

        rdopkg('patch', '-l')
        # after
        commit_after = git('rev-parse', 'HEAD')

        common.assert_spec_version(spec_version, spec_release_parts,
                                   spec_milestone)
        assert commit_before != commit_after, "No commit created"
        prev = git('rev-parse', 'HEAD~')
        assert prev == commit_before, "Multiple commits created"


@pytest.mark.skipif('RPM_AVAILABLE == False')
def test_patch_milestone(tmpdir):
    _test_patch('milestone', ('1.2.3', ('0.4', '%{?milestone}',
                DIST_POSTFIX), '.0rc2'), tmpdir)


@pytest.mark.skipif('RPM_AVAILABLE == False')
def test_patch_milestone_bug(tmpdir):
    # make sure rdopkg removes unwanted '%global milestone %{?milestone}'
    _test_patch('milestone-bug',
                ('1.2.3', ('0.4', '', DIST_POSTFIX), None),
                tmpdir)


@pytest.mark.skipif('RPM_AVAILABLE == False')
def test_patch_remove(tmpdir):

    dist_path = common.prep_spec_test(tmpdir, 'patched')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        common.add_patches()
        commit_before = git('rev-parse', 'HEAD')
        common.remove_patches(1)
        rdopkg('patch', '-l')
        commit_after = git('rev-parse', 'HEAD')
        git_clean = git.is_clean()
        common.norm_changelog()
    common.assert_distgit(dist_path, 'patch-remove')
    assert commit_before != commit_after, "New commit not created"
    assert git_clean, "git not clean after action"


@pytest.mark.skipif('RPM_AVAILABLE == False')
def test_patch_add(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'patched')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        common.add_patches()
        commit_before = git('rev-parse', 'HEAD')
        common.add_n_patches(3)
        rdopkg('patch', '-l')
        commit_after = git('rev-parse', 'HEAD')
        git_clean = git.is_clean()
        common.norm_changelog()
    common.assert_distgit(dist_path, 'patch-add')
    assert commit_before != commit_after, "New commit not created"
    assert git_clean, "git not clean after action"


@pytest.mark.skipif('RPM_AVAILABLE == False')
def test_patch_mix(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'patched')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        common.add_patches()
        commit_before = git('rev-parse', 'HEAD')
        common.remove_patches(1)
        common.add_n_patches(3)
        rdopkg('patch', '-l')
        commit_after = git('rev-parse', 'HEAD')
        git_clean = git.is_clean()
        common.norm_changelog()
    common.assert_distgit(dist_path, 'patch-mix')
    assert commit_before != commit_after, "New commit not created"
    assert git_clean, "git not clean after action"


def _test_patch_noop(tmpdir, distgit, cmd):
    dist_path = common.prep_spec_test(tmpdir, distgit)
    with dist_path.as_cwd():
        common.prep_patches_branch()
        common.add_patches()
        # regen patch files in order for hashes to match git
        rdopkg('update-patches', '--amend')
        commit_before = git('rev-parse', 'HEAD')
        rdopkg(*cmd)
        commit_after = git('rev-parse', 'HEAD')
        git_clean = git.is_clean()
    common.assert_distgit(dist_path, distgit)
    assert commit_before == commit_after, "New commit created for noop"
    assert git_clean, "git not clean after action"


def test_patch_noop(tmpdir):
    _test_patch_noop(tmpdir, 'patched', ['patch', '-l'])


def test_patch_noop_detect(tmpdir):
    _test_patch_noop(tmpdir,
                     'patched', ['patch', '-l', '--changelog', 'detect'])


def test_patch_noop_count(tmpdir):
    _test_patch_noop(tmpdir,
                     'patched', ['patch', '-l', '--changelog', 'count'])


def test_patch_noop_plain(tmpdir):
    _test_patch_noop(tmpdir, 'patched', ['patch', '-l', '-C', 'plain'])


def test_patch_noop_no_bump(tmpdir):
    _test_patch_noop(tmpdir, 'patched', ['patch', '-l', '--no-bump'])


def _test_patch_regen(tmpdir, distgit, distgit_after,
                      cmd, norm_changelog=True):
    dist_path = common.prep_spec_test(tmpdir, distgit)
    with dist_path.as_cwd():
        common.prep_patches_branch()
        common.add_patches()
        commit_before = git('rev-parse', 'HEAD')
        rdopkg(*cmd)
        commit_after = git('rev-parse', 'HEAD')
        git_clean = git.is_clean()
        if norm_changelog:
            common.norm_changelog()
    common.assert_distgit(dist_path, distgit_after)
    assert commit_before != commit_after, \
        "New commit not created after patch regen"
    assert git_clean, "git not clean after action"


def test_patch_regen(tmpdir):
    _test_patch_regen(tmpdir, 'patched', 'patched-regen', ['patch', '-l'])


def test_patch_regen_detect(tmpdir):
    _test_patch_regen(tmpdir, 'patched', 'patched-regen',
                      ['patch', '-l', '-C', 'detect'])


def test_patch_regen_count(tmpdir):
    _test_patch_regen(tmpdir, 'patched', 'patched-regen',
                      ['patch', '-l', '-C', 'count'])


def test_patch_regen_plain(tmpdir):
    _test_patch_regen(tmpdir, 'patched', 'patched-regen',
                      ['patch', '-l', '--changelog', 'plain'])


def test_patch_regen_no_bump(tmpdir):
    _test_patch_regen(tmpdir, 'patched', 'patched',
                      ['patch', '-l', '--no-bump'], norm_changelog=False)


def test_patch_unicode(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'patched')
    with dist_path.as_cwd():
        git('config', 'user.name', 'Přikrášlený Žluťoučký Kůň')
        common.prep_patches_branch()
        common.add_patches()
        commit_before = git('rev-parse', 'HEAD')
        with git_branch('master-patches'):
            common.do_patch('foofile', '#to chceš',
                            "Přikrášlený Žluťoučký Kůň")
            common.do_patch('foofile', '#to asi chceš', "Přikrášlení koně")
            common.do_patch('foofile', '#to nechceš', "ěščřžýáí")
        rdopkg('patch', '-l')
        commit_after = git('rev-parse', 'HEAD')
        git_clean = git.is_clean()
    assert commit_before != commit_after, "New commit not created"
    assert git_clean, "git not clean after action"



from rdopkg import guess
from collections import namedtuple
import pytest

VersionTestCase = namedtuple('VersionTestCase', ('expected', 'input_data'))


data_table_good = [
    VersionTestCase(('1.2.3', None), '1.2.3'),
    VersionTestCase(('1.2.3', 'vX.Y.Z'), 'v1.2.3'),
    VersionTestCase(('1.2.3', 'VX.Y.Z'), 'V1.2.3'),
    VersionTestCase(('banana', None), 'banana'),
]

data_table_bad = [
    VersionTestCase((None, None), None),
    VersionTestCase((None, None), []),
    VersionTestCase((None, None), ()),
    VersionTestCase((None, None), ''),
    VersionTestCase((None, None), {}),
]

data_table_ugly = [
    VersionTestCase((None, None), ('foo', 'bar', 'bah')),
    VersionTestCase((None, None), ['foo', 'bar', 'bah']),
    VersionTestCase((None, None), {'foo': 'bar'}),
]


def test_table_data_good_tag2version():
    for entry in data_table_good:
        assert entry.expected == guess.tag2version(entry.input_data)


def test_table_data_bad_tag2version():
    for entry in data_table_bad:
        # Input Validation should probably return to us (None, None)
        # assert entry.expected == guess.tag2version(entry.input_data)
        assert (entry.input_data, None) == guess.tag2version(entry.input_data)


def test_table_data_ugly_tag2version():
    for entry in data_table_ugly:
        # TODO: probably should be a more specific exception
        with pytest.raises(Exception):
            guess.tag2version(entry.input_data)


def test_version2tag_simple():
    assert '1.2.3' == guess.version2tag('1.2.3')


def test_version2tag_type1():
    assert 'v1.2.3' == guess.version2tag('1.2.3', 'vX.Y.Z')


def test_version2tag_type2():
    assert 'V1.2.3' == guess.version2tag('1.2.3', 'VX.Y.Z')



import shutil
import pytest

from rdopkg.actions.distgit.actions import update_patches
from rdopkg.utils.git import git
from rdopkg.utils import specfile
import rdopkg.exception

import test_common as common


def test_update_empty(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'empty')
    spec_path = dist_path.join('foo.spec')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        spec_before = spec_path.read()
        update_patches('master',
                       local_patches_branch='master-patches',
                       version='1.2.3')
        spec_after = spec_path.read()
    assert spec_after == spec_before
    with dist_path.as_cwd():
        common.add_patches()
        update_patches('master',
                       local_patches_branch='master-patches',
                       version='1.2.3')
        spec_after = spec_path.read()
    common.assert_distgit(dist_path, 'patched')


def test_update_noop(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'patched')
    spec_path = dist_path.join('foo.spec')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        spec_before = spec_path.read()
        common.add_patches()
        update_patches('master',
                       local_patches_branch='master-patches',
                       version='1.2.3')
        spec_after = spec_path.read()
        assert spec_after == spec_before
        spec_before = spec_after
        update_patches('master',
                       local_patches_branch='master-patches',
                       version='1.2.3')
        spec_after = spec_path.read()
        assert spec_after == spec_before


def test_update_full(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'some-ex')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        common.add_patches(extra=True)
        update_patches('master',
                       local_patches_branch='master-patches',
                       version='1.2.3')
    common.assert_distgit(dist_path, 'patched-ex')


def test_update_weird(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'empty-weird')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        common.add_patches(extra=True)
        update_patches('master',
                       local_patches_branch='master-patches',
                       version='1.2.3')
    common.assert_distgit(dist_path, 'patched-weird')


def test_update_dense(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'empty-dense')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        common.add_patches(extra=True)
        update_patches('master',
                       local_patches_branch='master-patches',
                       version='1.2.3')
    common.assert_distgit(dist_path, 'patched-dense')


def test_update_dense_ex(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'empty-dense-ex')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        common.add_patches(extra=True)
        update_patches('master',
                       local_patches_branch='master-patches',
                       version='1.2.3')
    common.assert_distgit(dist_path, 'patched-dense-ex')


def test_update_comments(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'some-comments')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        common.add_patches(extra=True)
        update_patches('master',
                       local_patches_branch='master-patches',
                       version='1.2.3')
    common.assert_distgit(dist_path, 'patched-comments')


def test_update_git_am(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'git-am')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        common.add_patches(extra=True)
        update_patches('master',
                       local_patches_branch='master-patches',
                       version='1.2.3')
        apply_method = specfile.Spec().patches_apply_method()
    assert apply_method == 'git-am'
    common.assert_distgit(dist_path, 'patched-git-am')


def test_update_git_am_buildarch_fail(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'git-am-fail')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        common.add_patches(extra=True)
        with pytest.raises(rdopkg.exception.LintProblemsFound):
            update_patches('master',
                           local_patches_branch='master-patches',
                           version='1.2.3')
        apply_method = specfile.Spec().patches_apply_method()
    assert apply_method == 'git-am'


def test_update_autosetup(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'autosetup')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        common.add_patches(extra=True)
        update_patches('master',
                       local_patches_branch='master-patches',
                       version='1.2.3')
        apply_method = specfile.Spec().patches_apply_method()
    assert apply_method == 'autosetup'
    common.assert_distgit(dist_path, 'patched-autosetup')


def test_wipe(tmpdir):
    shutil.copyfile(common.spec_path('some'), str(tmpdir.join('foo.spec')))
    with tmpdir.as_cwd():
        spec = specfile.Spec()
        spec.wipe_patches()
        spec.save()
        with open('foo.spec', 'r') as fp:
            spec_after = fp.read()
    with open(common.spec_path('empty'), 'r') as fp:
        spec_exp = fp.read()
    assert spec_after == spec_exp
    # second wipe must be a no-op
    with tmpdir.as_cwd():
        spec = specfile.Spec()
        spec.wipe_patches()
        spec.save()
        with open('foo.spec', 'r') as fp:
            spec_after = fp.read()
    assert spec_after == spec_exp


def test_filter_out(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'empty-ex-filter')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        common.add_patches(extra=True, filtered=True)
        update_patches('master',
                       local_patches_branch='master-patches',
                       version='1.2.3')
    common.assert_distgit(dist_path, 'patched-filter')


def test_update_milestone(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'milestone')
    spec_path = dist_path.join('foo.spec')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        spec_before = spec_path.read()
        update_patches('master',
                       local_patches_branch='master-patches',
                       version='1.2.3')
        spec_after = spec_path.read()
    assert spec_after == spec_before
    with dist_path.as_cwd():
        common.add_patches()
        update_patches('master',
                       local_patches_branch='master-patches',
                       version='1.2.3')
        spec_after = spec_path.read()
    common.assert_distgit(dist_path, 'patched-milestone')


def test_update_double_patches_base(tmpdir):
    dist_path = common.prep_spec_test(tmpdir, 'double-patches')
    with dist_path.as_cwd():
        common.prep_patches_branch()
        with pytest.raises(rdopkg.exception.LintProblemsFound):
            update_patches('master',
                           local_patches_branch='master-patches',
                           version='1.2.3')
    common.assert_distgit(dist_path, 'double-patches')



from rdopkg.action import Action, Arg


ACTIONS = [
    Action('reqdiff', help="show diff of requirements.txt",
           steps=[
               Action('get_package_env', module='distgit'),
               Action('get_diff_range', module='distgit'),
               Action('reqdiff'),
           ],
           optional_args=[
               Arg('diff_range', positional=True, nargs='*', metavar='GIT_REF',
                   help="no args: diff between current and upstream; "
                        "1 arg: diff between current and supplied git ref; "
                        "2 args: diff between 1st and 2nd supplied git refs"),
           ],
           ),
    Action('reqcheck',
           help="inspect requirements.txt vs .spec Requires",
           steps=[
               Action('get_package_env', module='distgit'),
               Action('reqcheck'),
           ],
           optional_args=[
               Arg('spec', shortcut='-s', action='store_true',
                   help="output .spec Requires: for easy pasting"),
           ]),
    Action('reqquery',
           help="query RDO repos for versions defined in requirements.txt",
           required_args=[
               Arg('filter', positional=True, metavar='RELEASE(/DIST)',
                   nargs='?',
                   help="RDO release(/dist) to query (see `rdopkg info`)"),
           ],
           optional_args=[
               Arg('reqs_file', shortcut='-r', metavar='FILE',
                   help="query supplied requirements.txt FILE"),
               Arg('reqs_ref', shortcut='-R', metavar='VERSION',
                   help="query requirements.txt from VERSION git ref"),
               Arg('spec', shortcut='-s', action='store_true',
                   help="query .spec file in current directory"),
               Arg('load', shortcut='-l', action='store_true',
                   help="load query results from requirements.yml "
                        "(created with -d)"),
               Arg('load_file', shortcut='-L', metavar='FILE',
                   help="load query results from FILE (created with -D)"),
               Arg('dump', shortcut='-d', action='store_true',
                   help="dump query results to requirements.yml "
                        "(view with -l)"),
               Arg('dump_file', shortcut='-D', metavar='FILE',
                   help="dump query results to FILE (view with -L)"),
               Arg('verbose', shortcut='-v', action='store_true',
                   help="print status during queries"),
           ]),
    Action('query',
           help="query RDO/dist repos for available package versions",
           optional_args=[
               Arg('filter', positional=True, metavar='RELEASE(/DIST)',
                   help="RDO release(/dist) to query (see `rdopkg info`)"),
               Arg('package', positional=True, metavar='PACKAGE',
                   help="package name to query about"),
               Arg('verbose', shortcut='-v', action='store_true',
                   help="print status during queries"),
           ]),
]



"""
Utility actions module.
"""

from rdopkg.conf import cfg, cfg_files
from rdopkg import exception
from rdopkg.actionmods import doctor as _doctor
from rdopkg.utils import log
from rdopkg import helpers


def status():
    raise exception.InternalAction(action='status')


def conf():
    if cfg_files:
        log.info("Following config files were read:")
        helpers.print_list(cfg_files)
    else:
        log.info("No rdopkg config files found, using default config:")
    log.info("")
    for item in cfg.items():
        log.info("%s: %s" % item)


def actions():
    raise exception.InternalAction(action='actions')


def autocomplete():
    try:
        import argcomplete  # NOQA
        print("argcomplete module is available.")
    except ImportError:
        print(
            "You're missing the argcomplete python module. "
            "Install it using\n\n"
            "    # dnf install -y python-argcomplete\n\n"
            "or\n\n"
            "    # pip install argcomplete\n\n"
            "and run `rdopkg autocomplete` again.")
        return
    print('')
    print(
        "Make sure following line is somewhere in your bash config:\n\n"
        '    eval "$(register-python-argcomplete rdopkg)"\n\n'
        "For zsh, you need bashcompinit, see https://github.com/kislyuk/"
        "argcomplete/issues/10#issuecomment-19369876")


def doctor():
    _doctor.can_haz_doctor()



from behave import *
import re
import os

from rdopkg.utils import distgitmagic
from rdopkg.utils import specfile
from rdopkg.utils.distgitmagic import git
from rdopkg.utils.testing import exdiff


def _create_distgit(context, version, release, magic_comments=None):
    name = 'foo-bar'
    context.execute_steps(u'Given a temporary directory')
    context.distgitdir = distgitmagic.create_sample_distgit(
        name, version=version, release=release, magic_comments=magic_comments)
    os.chdir(context.distgitdir)
    # collect .spec state to compare against after actions
    spec = specfile.Spec()
    context.old_changelog_entry = spec.get_last_changelog_entry()
    context.old_commit = git.current_commit()


@given('a distgit at Version {version} and Release {release} with magic comments')  # noqa
def step_impl(context, version, release):
    _create_distgit(context, version, release, magic_comments=context.text)


@given('a distgit at Version {version} and Release {release}')
def step_impl(context, version, release):
    _create_distgit(context, version, release)


@given('a distgit at Version {version}')
def step_impl(context, version):
    step = u'Given a distgit at Version %s and Release 2' % version
    context.execute_steps(step)


@given('a distgit')
def step_impl(context):
    context.execute_steps(u'Given a distgit at Version 1.2.3 and Release 2')


@given('a distgit with Change-Id {changeid}')
def step_impl(context, changeid):
    context.execute_steps(u'Given a distgit at Version 1.2.3 and Release 2')
    git('commit', '--amend', '-m',
        context.old_commit + '\n\nChange-Id: %s' % changeid)
    context.old_commit = git.current_commit()


@given('a patches branch with {n:n} patches')
def step_impl(context, n):
    distgitmagic.create_sample_patches_branch(n)


@given('a patches branch with following patches')
def step_impl(context):
    patches = context.text.splitlines()
    distgitmagic.create_sample_patches_branch(patches=patches)


@given('a patches branch with {n:n} patches without version git tag')
def step_impl(context, n):
    distgitmagic.create_sample_patches_branch(n, version_tag=False)


@given('a new version {version} with {n:n} patches from patches branch')
def step_impl(context, version, n):
    distgitmagic.create_sample_upstream_new_version(version, 9, n)


@given('a new version {version}')
def step_impl(context, version):
    context.execute_steps(
        u'Given a new version %s with 0 patches from patches branch' %
        version)


@given(u'a local file {fn} containing "{text}"')
def step_impl(context, fn, text):
    with open(os.path.join(context.distgitdir, fn), 'w') as f:
        f.write(text)


@given(u'a local file {fn}')
def step_impl(context, fn):
    with open(os.path.join(context.distgitdir, fn), 'w') as f:
        f.write(context.text)


@when(u'I set .spec file patches_base={base}')
def step_impl(context, base):
    spec = specfile.Spec()
    spec.set_patches_base(base)
    spec.save()


@when(u'I set .spec file patches_base to existing commit +{n:n}')
def step_impl(context, n):
    pb = git('show-ref', 'master-patches')[:8]
    if n:
        pb = '%s+%s' % (pb, n)
    context.execute_steps(u'When i set .spec file patches_base=%s' % pb)


@when('I change .spec file tag {tag} to {value}')
def step_impl(context, tag, value):
    spec = specfile.Spec()
    spec.set_tag(tag, value)
    spec.save()


@when(u'I prepend .spec file with')
def step_impl(context):
    spec = specfile.Spec()
    spec._txt = context.text + spec.txt
    spec.save()


@when(u'I undo all changes')
def step_impl(context):
    git("stash")
    assert git.is_clean()


@when('I add description to .spec changelog')
def step_impl(context):
    spec = specfile.Spec()
    spec._txt, n = re.subn('(\n%changelog\n\*[^\n]+\n)\n',
                           '\g<1>- Description of a change\n',
                           spec.txt)
    assert n == 1
    spec.save()


@then('.spec file tag {tag} is {value}')
def step_impl(context, tag, value):
    spec = specfile.Spec()
    assert spec.get_tag(tag) == value, \
        "{0} != {1}".format(spec.get_tag(tag), value)


@then('.spec file contains new changelog entry with {n:n} lines')
def step_impl(context, n):
    spec = specfile.Spec()
    entry = spec.get_last_changelog_entry()
    assert len(entry[1]) == n
    assert entry != context.old_changelog_entry


@then('.spec file doesn\'t contain new changelog entries')
def step_impl(context):
    entry = specfile.Spec().get_last_changelog_entry()
    assert entry == context.old_changelog_entry


@then('.spec file contains new changelog entry with {text}')
def step_impl(context, text):
    spec = specfile.Spec()
    entry = spec.get_last_changelog_entry()
    changelog_block = '\n'.join(entry[1])
    assert text in changelog_block, \
        "[{0}] not found in [{1}]".format(text, changelog_block)
    assert entry != context.old_changelog_entry


@then('.spec file has {n:n} patches defined')
def step_impl(context, n):
    spec = specfile.Spec()
    assert spec.get_n_patches() == n


@then('.spec file doesn\'t contain patches_base')
def step_impl(context):
    spec = specfile.Spec()
    assert spec.get_patches_base() == (None, 0)


@then('.spec file contains /{rex}/')
def step_impl(context, rex):
    spec = specfile.Spec()
    assert re.search(rex, spec.txt), "/%s/ not found in .spec" % rex


@then('.spec file contains {text}')
def step_impl(context, text):
    spec = specfile.Spec()
    assert text in spec.txt, "%s not found in .spec" % text


@then('new commit was created')
def step_impl(context):
    new_commit = git.current_commit()
    assert new_commit != context.old_commit


@then(u'no new commit was created')
def step_impl(context):
    new_commit = git.current_commit()
    assert new_commit == context.old_commit


@then(u'last commit message contains {simple_string}')
def step_impl(context, simple_string):
    msg = git.current_commit_message()
    assert simple_string in msg, \
        (u"'{0}' not found in:\n{1}".format(simple_string, msg)
         ).encode('ascii', 'replace')


@then(u'last commit message is')
def step_impl(context):
    msg = git.current_commit_message()
    assert context.text == msg, exdiff(
        context.text, msg,
        header="Commit message differs from expected format:")


@then(u'git is clean')
def step_impl(context):
    assert git.is_clean(), git('status').encode('ascii', 'replace')



class RdopkgException(Exception):
    msg_fmt = "An unknown error occurred"
    exit_code = 1

    def __init__(self, msg=None, exit_code=None, **kwargs):
        self.kwargs = kwargs
        if not msg:
            try:
                msg = self.msg_fmt % kwargs
            except Exception:
                # kwargs doesn't match those in message.
                # Returning this is still better than nothing.
                msg = self.msg_fmt
        if exit_code is not None:
            self.exit_code = exit_code
        super(RdopkgException, self).__init__(msg)


class CommandFailed(RdopkgException):
    msg_fmt = "Command failed: %(cmd)s"


class CommandNotFound(RdopkgException):
    msg_fmt = "Command not found: %(cmd)s"


class InvalidRemoteBranch(RdopkgException):
    msg_fmt = "Git remote not found for remote branch: %(branch)s"


class BranchNotFound(RdopkgException):
    msg_fmt = "Expected git branch not found: %(branch)s"


class SpecFileNotFound(RdopkgException):
    msg_fmt = "No .spec files found."


class IncompleteChangelog(RdopkgException):
    msg_fmt = "Description of changes is missing in %%changelog."


class MultipleChangelog(RdopkgException):
    msg_fmt = "Multiple %%changelog sections found. Expected only one."


class MultipleSpecFilesFound(RdopkgException):
    msg_fmt = "Multiple .spec files found. Expected only one."


class SpecFileParseError(RdopkgException):
    msg_fmt = "Error parsing .spec file '%(spec_fn)s': %(error)s"


class ModuleNotAvailable(RdopkgException):
    msg_fmt = "Module %(module)s is not available. Unable to continue."


class RpmModuleNotAvailable(ModuleNotAvailable):
    msg_fmt = ("Module rpm is not available. It is required to parse .spec "
               "files. Pro tip: `dnf install rpm-python`")


class InvalidAction(RdopkgException):
    msg_fmt = "Invalid action: %(action)s"


class InvalidActionModule(RdopkgException):
    msg_fmt = "Invalid action module: %(why)s"


class DuplicateAction(RdopkgException):
    msg_fmt = "Duplicate action '%(action)s' in modules: %(mod1)s, %(mod2)s"


class InvalidGitRef(RdopkgException):
    msg_fmt = "Invalid git reference: %(ref)s"


class InvalidPatchesBaseRef(RdopkgException):
    msg_fmt = ("Invalid base patches branch git reference: %(ref)s\n\n"
               "Make sure .spec Version references existing git tag\n"
               "or select explicitly using #patches_base=ANY_GIT_REF\n\n"
               "`rdopkg pkgenv` can help you debug this!")


class UserAbort(RdopkgException):
    msg_fmt = "Aborted by user."


class NoActionInProgress(RdopkgException):
    msg_fmt = "No action in progress."


class ActionInProgress(RdopkgException):
    msg_fmt = "Please `--continue` or `--abort` previous action before " \
              "running a new one."


class ActionFunctionNotAvailable(RdopkgException):
    msg_fmt = "Action function not available: %(module)s:%(action)s"


class RequiredActionArgumentNotAvailable(RdopkgException):
    msg_fmt = "Required argument of '%(action)s' action not available: %(arg)s"


class InvalidUsage(RdopkgException):
    msg_fmt = "Invalid usage: %(why)s"


class InvalidQuery(RdopkgException):
    msg_fmt = "Invalid query: %(why)s"


class InvalidPackage(RdopkgException):
    msg_fmt = "Package not found in distroinfo: %(package)s"


class InvalidPackageFilter(RdopkgException):
    msg_fmt = "Invalid package filter: %(why)s"


class InvalidReleaseBumpIndex(RdopkgException):
    msg_fmt = "Invalid Release bump index: %(what)s"


class CantGuess(RdopkgException):
    msg_fmt = "Unable to determine %(what)s: %(why)s"


class ManualResolutionNeeded(RdopkgException):
    msg_fmt = "Your intervention is required. Run `rdopkg -c` again when done."


class ConfigError(RdopkgException):
    msg_fmt = "Configuration error: %(what)s"


class RepoError(RdopkgException):
    msg_fmt = "Repository error: %(what)s"


class CommandOutputParseError(RdopkgException):
    msg_fmt = "Failed to parse %(tool)s output:\n%(output)s"


class FileNotFound(RdopkgException):
    msg_fmt = "File not found: %(path)s"


class OnlyPatchesIgnoreUsed(RdopkgException):
    msg_fmt = ("update-patches attempted with only #patches_ignore magic "
               "comment in the spec file. Currently it is not safe to run "
               "update-patches without #patches_base specified AFTER "
               "#patches_ignore.")


class NotADirectory(RdopkgException):
    msg_fmt = "Not a directory: %(path)s"


class NotAFile(RdopkgException):
    msg_fmt = "Not a file: %(path)s"


class NewPackageAlreadyPresent(RdopkgException):
    msg_fmt = "Can't copy package, file already exists: %(path)s"


class ToolNotFound(RdopkgException):
    msg_fmt = "%(tool)s not found in %(path)s and it's parents."


class ActionRequired(RdopkgException):
    msg_fmt = "Action required"


class ActionFinished(RdopkgException):
    msg_fmt = "Action finished"


class ActionGoto(RdopkgException):
    msg_fmt = "Jump to action: %(goto)s"


class DebugAbort(RdopkgException):
    msg_fmt = "DEBUG abort"


class RequirementNotMet(RdopkgException):
    msg_fmt = "Requirement not met: %(what)s"


class InternalAction(RdopkgException):
    msg_fmt = "Unhandled internal action: %(action)s"


class UnverifiedPatch(RdopkgException):
    msg_fmt = "Patch not validated by CI, use --force to override"


class NoPatchesChanged(RdopkgException):
    msg_fmt = "No patches changed"


class NoDistgitChangesFound(RdopkgException):
    msg_fmt = "No distgit changes found"


class InvalidLintCheck(RdopkgException):
    msg_fmt = "Invalid lint check: %(check)s"


class LintProblemsFound(RdopkgException):
    exit_code = 23
    msg_fmt = "Lint problems detected."



from __future__ import unicode_literals

import os
import py
import re
import shutil

from rdopkg.utils.cmd import encode
from rdopkg.utils.git import git
from rdopkg.utils.specfile import Spec
from rdopkg import exception


ASSETS_DIR = 'tests/assets'
DIST_POSTFIX = '%{?dist}'


def ffind(path='.', include_dirs=True):
    tree = set()
    for root, dirs, files in os.walk(path):
        if include_dirs:
            for f in dirs:
                tree.add(os.path.join(root, f))
        for f in files:
            tree.add(os.path.join(root, f))
    return tree


def cfind(path, include_dirs=True):
    if not hasattr(path, 'as_cwd'):
        path = py.path.local(path=path)
    with path.as_cwd():
        t = ffind(include_dirs=include_dirs)
    t = set(filter(lambda x: not x.startswith('./.'), t))
    return t


def distgit_path(dg):
    return os.path.join(ASSETS_DIR, 'spec', dg)


def spec_path(dg):
    return os.path.join(distgit_path(dg), 'foo.spec')


def prep_spec_test(tmpdir, distgit):
    dist_path = tmpdir.join('dist')
    shutil.copytree(os.path.join(ASSETS_DIR, 'spec', distgit),
                    str(dist_path))
    with dist_path.as_cwd():
        git('init')
        git('add', '.')
        git('commit', '-m', 'Initial import',
            isolated=True)
    return dist_path


def prep_patches_branch(tag='1.2.3'):
    git('checkout', '--orphan', 'master-patches')
    f = open('foofile', 'w')
    f.write("#not really a patch\n")
    f.close()
    git('add', 'foofile')
    git('commit', '-m', 'Create this test branch',
        isolated=True)
    if tag:
        git('tag', tag)
    git('checkout', 'master')


def do_patch(fn, content, msg):
    f = open(fn, 'w')
    content = encode(content)
    f.write(content)
    f.close()
    git('add', fn)
    git('commit', '-m', msg,
        isolated=True)


def add_patches(extra=False, filtered=False, tag=None):
    git('checkout', 'master-patches')
    if extra:
        do_patch('foofile', "#meh\n", 'Look, excluded patch')
        do_patch('foofile', "#nope\n", 'Yet another excluded patch')
    do_patch('foofile', "#huehue, change\n", 'Crazy first patch')
    if filtered:
        do_patch('foofile', "#fix ci\n", 'DROP-IN-RPM: ci fix')
        do_patch('foofile', "#and now for real\n", 'DROP-IN-RPM: moar ci fix')
    do_patch('foofile', "#lol, another change\n", 'Epic bugfix of doom MK2')
    if filtered:
        do_patch('foofile', "#oooops\n", 'DROP-IN-RPM: even moar ci fix')
    if tag:
        git('tag', tag)
    git('checkout', 'master')


def add_n_patches(n):
    git('checkout', 'master-patches')
    for i in range(1, n + 1):
        do_patch('foofile', "#extra patch %d\n" % i, 'Extra patch %d' % i)
    git('checkout', 'master')


def remove_patches(n):
    git('checkout', 'master-patches')
    git('reset', '--hard', 'HEAD' + n * '~')
    git('checkout', 'master')


def assert_distgit(dg_path, img_dg):
    img_path = distgit_path(img_dg)
    current_tree = cfind(dg_path)
    exp_tree = cfind(img_path)
    assert current_tree == exp_tree
    current_txt = dg_path.join('foo.spec').read()
    with open(os.path.join(img_path, 'foo.spec')) as fp:
        exp_txt = fp.read()
    assert current_txt == exp_txt


def assert_spec_version(version, release_parts, milestone):
    spec = Spec()
    spec_version = spec.get_tag('Version')
    spec_release_parts = spec.get_release_parts()
    spec_milestone = spec.get_milestone()

    assert spec_version == version
    assert spec_release_parts == release_parts
    assert spec_milestone == milestone


def norm_changelog(count=1):
    spec = Spec()
    r = re.compile(r"%changelog\n", flags=re.I).split(spec.txt)
    if len(r) > 2:
        raise exception.MultipleChangelog()
    txt, chl = r
    chl, n = re.subn(r'^\* .+\s(\d\S*)$',
                     '* DATE AUTHOR \g<1>',
                     chl,
                     count=count,
                     flags=re.M)
    spec._txt = txt + '%changelog\n' + chl
    spec.save()



from rdopkg.action import Action, Arg


ACTIONS = [
    Action('clone',
           help="custom clone action replacing internal one",
           required_args=[
               Arg('package', positional=True, metavar='PACKAGE',
                   help="custom package to clone"),
           ]),
]



# -*- encoding: utf-8 -*-

import random
import re

from rdopkg.utils import log

# replace red with another green for more positive experience
COLORS = ['green', 'green', 'blue', 'yellow', 'magenta', 'cyan', 'white']
WISDOM = [
    u"We are the universe observing itself.",
    u"Not having any problems is a problem.",
    u"Deriving meaning from words is a great pleasure.",
    u"There is always more work. ¯\_(ツ)_/¯",
    u"rdopkg likes you ;)",
    u"Don't take life too seriously.",
    u"rdopkg is here for you *^^*",
    u"When you're the smartest person in a room, you're in the wrong room.",
    u"Do what you want cause a pirate is free and you are a pirate.",
    u"Free and Open Source Software has won the war \o/",
    u"We've learned as much as we have suffered.",
    u"Bend the universe to your will using words.",
    u"At least you're not writing Visual Basic.",
    u"Using words we command machines that run the world.",
    u"Cooperation is hard to establish and hard to maintain. Don't give up.",
    u"Work will become obsolete as soon as the singularity is born.",
    u"Life is not a zero-sum game.",
    u"Never give up, never surrender!",
    u"Apply critical thinking.",
    u"Construct Additional Pylons!",
    u"Have you tried turning it off and on again?",
    u"Minimalism is the answer.",
    u"Firing orbital friendship cannon on this location.",
    (u"Humans put everything into boxes because it's easier that way. "
     u"Packaging is important."),
    (u"Language is the most powerful tool known to humankind. "
     u"Don't you enjoy using it?"),
    (u"Scientists explore the universe as it is, "
     u"engineers create the universe that has never been."),
]


def royal_rainbow(lulz):
    s = '\n'
    for lul in re.split(r'\s+', lulz):
        s += '{t.%s}%s ' % (random.choice(COLORS), lul)
    s += '\n'
    print(s.format(t=log.term))


def can_haz_doctor():
    advice = random.choice(WISDOM)
    royal_rainbow(advice)



from rdopkg.cli import rdopkg
from rdopkg.utils.git import git
from rdopkg.utils import log

import test_common as common
from test_common import DIST_POSTFIX
import pytest

RPM_AVAILABLE = False
try:
    import rpm  # NOQA
    RPM_AVAILABLE = True
except ImportError:
    pass


def _test_new_version(asset, dir, steps):
    dist_path = common.prep_spec_test(dir, asset)
    log.log.setLevel(log.WARN)

    with dist_path.as_cwd():
        common.prep_patches_branch()
        for new_version, spec_version, spec_release_parts, spec_milestone \
                in steps:
            commit_before = git('rev-parse', 'HEAD')
            common.add_patches(tag=new_version)
            rdopkg('new-version', '-l', '-d', new_version)

            # after
            commit_after = git('rev-parse', 'HEAD')
            common.assert_spec_version(spec_version, spec_release_parts,
                                       spec_milestone)
            assert commit_before != commit_after


@pytest.mark.skipif('RPM_AVAILABLE == False')
def test_new_version_basic(tmpdir):
    steps = [
        ('3.4.5', '3.4.5', ('1', '', DIST_POSTFIX), None),
        ('123.456.789.000', '123.456.789.000', ('1', '', DIST_POSTFIX), None),
    ]
    _test_new_version('some', tmpdir, steps)


@pytest.mark.skipif('RPM_AVAILABLE == False')
def test_new_version_milestone(tmpdir):
    steps = [
        ('3.4.5', '3.4.5', ('1', '', DIST_POSTFIX), None),
        ('4.0.0.0b1',
         '4.0.0', ('0.1', '%{?milestone}', DIST_POSTFIX), '.0b1'),
        ('4.0.0.0rc1',
         '4.0.0', ('0.2', '%{?milestone}', DIST_POSTFIX), '.0rc1'),
        ('4.0.0', '4.0.0', ('1', '', DIST_POSTFIX), None),
    ]
    _test_new_version('some', tmpdir, steps)



from rdopkg.action import Action, Arg


ACTIONS = [
    Action('info', help="show information about RDO packaging",
           optional_args=[
               Arg('pkgs', positional=True, nargs='*', metavar='ATTR:REGEX',
                   help="show info about packages with ATTR matching REGEX"),
               Arg('apply_tag', shortcut='-t',
                   help="apply overrides for selected tag"),
               Arg('force_fetch', shortcut='-f', action='store_true',
                   help="force fetch of info repo"),
               Arg('local_info', shortcut='-l',
                   help="use local distroinfo repo found in specified path"),
               Arg('info_file', shortcut='-i',
                   help="use specified distroinfo info file"),
           ]),
    Action('info_tags_diff',
           help=("find which tags have changed between HEAD~..HEAD"
                 " in distroinfo"),
           optional_args=[
               Arg('local_info', positional=True, metavar='INFO_REPO',
                   help="use local distroinfo repo found in INFO_REPO"),
               Arg('info_file', positional=True, metavar='INFO_FILE',
                   nargs='?',
                   help="use specified distroinfo INFO_FILE"),
               Arg('buildsys_tags', shortcut='-b', action='store_true',
                   help="process buildsys-tags instead of regular tags"),
           ]),
    Action('findpkg',
           help="find and show single best matching package in distroinfo",
           optional_args=[
               Arg('query', positional=True, metavar='PACKAGE/PROJECT/URL',
                   help="project name, package name or upstream URL"),
               Arg('strict', shortcut='-s', action='store_true',
                   help="only match whole pkg/proj/URL, no substring magics"),
               Arg('force_fetch', shortcut='-f', action='store_true',
                   help="force fetch of distroinfo repo"),
               Arg('local_info', shortcut='-l',
                   help="use local distroinfo repo found in specified path"),
               Arg('info_file', shortcut='-i',
                   help="use specified distroinfo info file"),
           ]),
]



from __future__ import unicode_literals

import codecs
from collections import defaultdict
import os
import re
import time

from rdopkg import exception
from rdopkg.utils import lint

RPM_AVAILABLE = False
try:
    import rpm
    RPM_AVAILABLE = True
except ImportError:
    pass


RELEASE_PARTS_SEMVER = {
    'MAJOR': 1,
    'MINOR': 2,
    'PATCH': 3,
}


def split_filename(filename):
    """
    Received a standard style rpm fullname and returns
    name, version, release, epoch, arch
    Example: foo-1.0-1.i386.rpm returns foo, 1.0, 1, i386
             1:bar-9-123a.ia64.rpm returns bar, 9, 123a, 1, ia64

    This function replaces rpmUtils.miscutils.splitFilename, see
    https://bugzilla.redhat.com/1452801
    """

    # Remove .rpm suffix
    if filename.endswith('.rpm'):
        filename = filename.split('.rpm')[0]

    # is there an epoch?
    components = filename.split(':')
    if len(components) > 1:
        epoch = components[0]
    else:
        epoch = ''

    # Arch is the last item after .
    arch = filename.rsplit('.')[-1]
    remaining = filename.rsplit('.%s' % arch)[0]
    release = remaining.rsplit('-')[-1]
    version = remaining.rsplit('-')[-2]
    name = '-'.join(remaining.rsplit('-')[:-2])

    return name, version, release, epoch, arch


def string_to_version(verstring):
    """
    Return a tuple of (epoch, version, release) from a version string

    This function replaces rpmUtils.miscutils.stringToVersion, see
    https://bugzilla.redhat.com/1364504
    """
    # is there an epoch?
    components = verstring.split(':')
    if len(components) > 1:
        epoch = components[0]
    else:
        epoch = 0

    remaining = components[:2][0].split('-')
    version = remaining[0]
    release = remaining[1]

    return (epoch, version, release)


def spec_fn(spec_dir='.'):
    """
    Return the filename for a .spec file in this directory.
    """
    specs = [f for f in os.listdir(spec_dir)
             if os.path.isfile(f) and f.endswith('.spec')]
    if not specs:
        raise exception.SpecFileNotFound()
    if len(specs) != 1:
        raise exception.MultipleSpecFilesFound()
    return specs[0]


def get_patches_from_files(patches_dir='.'):
    patches_fns = [f for f in os.listdir(patches_dir)
                   if os.path.isfile(f) and f.endswith('.patch')]
    if not patches_fns:
        return []
    patches = []
    for pfn in patches_fns:
        with codecs.open(pfn, 'r', encoding='utf-8') as fp:
            txt = fp.read()
        hash = None
        m = re.search(r'^From ([a-z0-9]+)', txt, flags=re.M)
        if m:
            hash = m.group(1)
        subj = None
        m = re.search(r'^Subject:\w*(.+)$', txt, flags=re.M)
        if m:
            subj = m.group(1)
        patches.append((pfn, hash, subj))
    return patches


def version_parts(version):
    """
    Split a version string into numeric X.Y.Z part and the rest (milestone).
    """
    m = re.match(r'(\d+(?:\.\d+)*)([.%]|$)(.*)', version)
    if m:
        numver = m.group(1)
        rest = m.group(2) + m.group(3)
        return numver, rest
    else:
        return version, ''


def release_parts(version):
    """
    Split RPM Release string into (numeric X.Y.Z part, milestone, rest).

    :returns: a three-element tuple (number, milestone, rest). If we cannot
              determine the "milestone" or "rest", those will be an empty
              string.
    """
    numver, tail = version_parts(version)
    if numver and not re.match(r'\d', numver):
        # entire release is macro a la %{release}
        tail = numver
        numver = ''
    m = re.match(r'(\.?(?:%\{\?milestone\}|[^%.]+))(.*)$', tail)
    if m:
        milestone = m.group(1)
        rest = m.group(2)
    else:
        milestone = ''
        rest = tail
    return numver, milestone, rest


def has_macros(s):
    # detect escaping (%%)
    rex = r'.*(?<!%)%[\w{].*'
    if re.match(rex, s):
        return True
    return False


def nvrcmp(nvr1, nvr2):
    if not RPM_AVAILABLE:
        raise exception.RpmModuleNotAvailable()
    t1 = string_to_version(nvr1)
    t2 = string_to_version(nvr2)
    return rpm.labelCompare(t1, t2)


def vcmp(v1, v2):
    if not RPM_AVAILABLE:
        raise exception.RpmModuleNotAvailable()
    t1 = ('0', v1, '')
    t2 = ('0', v2, '')
    return rpm.labelCompare(t1, t2)


def nvr2version(nvr):
    _, v, _, _, _ = split_filename(nvr)
    return v


class Spec(object):
    """
    Lazy .spec file parser and editor.
    """

    RE_PATCH = r'(?:^|\n)(Patch\d+:)'
    RE_AFTER_SOURCES = r'((?:^|\n)Source\d*:[^\n]*\n\n?)'
    RE_AFTER_PATCHES_BASE = (
        r'((?:^|\n)(?:#[ \t]*\n)*#\s*patches_base\s*=[^\n]*\n(?:#[ '
        r'\t]*\n)*)\n*')
    RE_MACRO_BASE = r'%global\s+{0}\s+'

    def __init__(self, fn=None, txt=None):
        """
        Spec file reader/writer/parser.

        :param  fn: The filename of a .spec file. If not provided, we will
                    select the only .spec file in current directory or throw an
                    exception when multiple or no .spec files are present.
        :type   fn: ``str``

        :param txt: The textual contents of a .spec file. If not provided, we
                    will read the contents from disk.
        :type  txt: ``str``
        """
        self._fn = fn
        self._txt = txt
        self._rpmspec = None

    @property
    def fn(self):
        """ The filename of this .spec file. """
        if not self._fn:
            self._fn = spec_fn()
        return self._fn

    @property
    def txt(self):
        """ The textual contents of this .spec file. """
        if not self._txt:
            with codecs.open(self.fn, 'r', encoding='utf-8') as fp:
                self._txt = fp.read()
        return self._txt

    def load_rpmspec(self):
        if not RPM_AVAILABLE:
            raise exception.RpmModuleNotAvailable()
        rpm.addMacro('_sourcedir',
                     os.path.dirname(os.path.realpath(self.fn)))
        try:
            self._rpmspec = rpm.spec(self.fn)
        except ValueError as e:
            raise exception.SpecFileParseError(spec_fn=self.fn,
                                               error=e.args[0])

    @property
    def rpmspec(self):
        if not self._rpmspec:
            self.load_rpmspec()
        return self._rpmspec

    def expand_macro(self, macro):
        if not self._rpmspec:
            self.load_rpmspec()
        if not RPM_AVAILABLE:
            raise exception.RpmModuleNotAvailable()
        return rpm.expandMacro(macro)

    def get_tag(self, tag, default=exception.SpecFileParseError,
                expand_macros=False):
        m = re.search(r'^%s:\s+(\S.*)$' % tag, self.txt, re.M)
        if not m:
            if default != exception.SpecFileParseError:
                return default
            raise exception.SpecFileParseError(spec_fn=self.fn,
                                               error="%s tag not found" % tag)
        tag = m.group(1).rstrip()
        if expand_macros and has_macros(tag):
            # don't parse using rpm unless required
            tag = self.expand_macro(tag)
        return tag

    def set_tag(self, tag, value):
        self._txt, n = re.subn(r'^(%s:\s+).*$' % re.escape(tag),
                               r'\g<1>%s' % value, self.txt, flags=re.M)
        return n > 0

    def get_tag_align_ws(self, tag):
        if not tag.endswith(':'):
            tag += ':'
        m = re.search(r'^%s(\s*)' % re.escape(tag), self.txt, flags=re.M)
        if not m:
            return ''
        return m.group(1)

    def get_magic_comment(self, name, expand_macros=False):
        """Return a value of # name=value comment in spec or None."""
        match = re.search(r'^#\s*?%s\s?=\s?(\S+)' % re.escape(name),
                          self.txt, flags=re.M)
        if not match:
            return None

        val = match.group(1)
        if expand_macros and has_macros(val):
            # don't parse using rpm unless required
            val = self.expand_macro(val)
        return val

    def get_patches_base(self, expand_macros=False):
        """Return a tuple (version, number_of_commits) that are parsed
        from the patches_base in the specfile.
        """
        match = re.search(r'(?<=patches_base=)[\w.+?%{}]+', self.txt)
        if not match:
            return None, 0

        patches_base = match.group()
        if expand_macros and has_macros(patches_base):
            # don't parse using rpm unless required
            patches_base = self.expand_macro(patches_base)
        patches_base_ref, _, n_commits = patches_base.partition('+')

        try:
            n_commits = int(n_commits)
        except ValueError:
            n_commits = 0
        return patches_base_ref, n_commits

    def get_patches_ignore_regex(self):
        """Returns a string representing a regex for filtering out patches

        This string is parsed from a comment in the specfile that contains the
        word filter-out followed by an equal sign.

        For example, a comment as such:
            # patches_ignore=(regex)

        would mean this method returns the string '(regex)'

        Only a very limited subset of characters are accepted so no fancy stuff
        like matching groups etc.
        """
        match = re.search(r'# *patches_ignore=([\w *.+?[\]|{,}\-_]+)',
                          self.txt)
        if not match:
            return None
        regex_string = match.group(1)
        try:
            return re.compile(regex_string)
        except Exception:
            return None

    def _create_new_patches_base(self, base):
        self._txt, n = re.subn(
            self.RE_PATCH,
            r'\n#\n# patches_base=%s\n#\n\g<1>' % base,
            self.txt, count=1, flags=re.M)
        if n != 1:
            self._txt, n = re.subn(
                self.RE_AFTER_SOURCES,
                r'\g<1>#\n# patches_base=%s\n#\n\n' % base,
                self.txt, count=1, flags=re.M)
        if n != 1:
            raise exception.SpecFileParseError(
                spec_fn=self.fn,
                error="Unable to create new #patches_base entry.")

    def set_patches_base(self, base):
        v, _ = self.get_patches_base()

        if re.search(r'^#\s*patches_ignore\s*=\s*\S+', self.txt, flags=re.M):
            # This is a temporary hack as patches_ignore currently requires
            # explicit patches_base. This should be solved with a proper
            # magic comment parser and using Version in filtration logic
            # when no patches_base is defined.
            if not base:
                base = self.get_tag('Version', expand_macros=True)
        if base:
            regex = r'^#\s*patches_base*'
            if v is None and re.search(regex, self.txt, flags=re.M) is None:
                self._create_new_patches_base(base)
            else:
                lines = self.txt.split('\n')
                patch_base_regex = re.compile(r'(#\s*patches_base\s*=\s*)\w*')
                for idx, line in enumerate(lines):
                    match = patch_base_regex.match(line)
                    if match is not None:
                        out_str = '{0}{1}'.format(match.group(1), base)
                        lines[idx] = out_str
                        break
                else:
                    raise exception.SpecFileParseError(
                        spec_fn=self.fn,
                        error="Unable to set new #patches_base")
                self._txt = '\n'.join(lines)

        else:
            if v is not None:
                # Drop magic comment patches_base and following empty comments
                self._txt = re.sub(
                    r'(?:^#)+\s*patches_base\s*=[^\n]*\n(?:^#\n)*',
                    '', self.txt, flags=re.M)

    def set_patches_base_version(self, version, ignore_macros=True):
        if not version:
            version = ''
        old_pb, n_commits = self.get_patches_base()
        if (ignore_macros and old_pb and has_macros(old_pb)):
            return False
        if n_commits > 0:
            version += ("+%s" % n_commits)
        self.set_patches_base(version)
        return True

    def get_n_patches(self):
        return len(re.findall(r'^Patch[0-9]+:', self.txt, re.M))

    def get_n_excluded_patches(self):
        """
        Gets number of excluded patches from patches_base:
        #patches_base=1.0.0+THIS_NUMBER
        """
        _, n_commits = self.get_patches_base()
        return n_commits

    def get_patch_fns(self):
        fns = []
        for m in re.finditer(r'^\s*Patch\d+:\s*(\S+)\s*$', self.txt,
                             flags=re.M):
            fns.append(m.group(1))
        return fns

    def wipe_patches(self):
        self._txt = re.sub(r'\n+(?:(?:Patch|.patch)\d+[^\n]*)', '', self.txt)

    def sanity_check(self):
        hints = lint.lint(self.fn, checks=['sanity'])
        lint.lint_report(hints, error_level='E')

    def patches_apply_method(self):
        if '\ngit am %{patches}' in self.txt:
            return 'git-am'
        if '\n%autosetup' in self.txt:
            return 'autosetup'
        return 'rpm'

    def set_commit_ref_macro(self, ref):
        self._txt = re.sub(
            r'^\%global commit \w+',
            '%%global commit %s' % ref, self.txt, flags=re.M)

    def set_new_patches(self, fns):
        self.wipe_patches()
        if not fns:
            return
        apply_method = self.patches_apply_method()
        ps = ''
        pa = ''
        for i, pfn in enumerate(fns, start=1):
            ps += "Patch%04d: %s\n" % (i, pfn)
            if apply_method == 'rpm':
                pa += "%%patch%04d -p1\n" % i
        # PatchXXX: lines after Source0 / #patches_base=
        self._txt, n = re.subn(
            self.RE_AFTER_PATCHES_BASE,
            r'\g<1>%s\n' % ps, self.txt, count=1)

        if n != 1:
            m = None
            for m in re.finditer(self.RE_AFTER_SOURCES, self.txt):
                pass
            if not m:
                raise exception.SpecFileParseError(
                    spec_fn=self.fn,
                    error="Failed to append PatchXXXX: lines")
            i = m.end()
            startnl, endnl = '', ''
            if self._txt[i - 2] != '\n':
                startnl += '\n'
            if self._txt[i] != '\n':
                endnl += '\n'
            self._txt = self._txt[:i] + startnl + ps + endnl + self._txt[i:]
        # %patchXXX -p1 lines after "%setup" if needed
        if apply_method == 'rpm':
            self._txt, n = re.subn(
                r'((?:^|\n)%setup[^\n]*\n)\s*',
                r'\g<1>\n%s\n' % pa, self.txt)
            if n == 0:
                raise exception.SpecFileParseError(
                    spec_fn=self.fn,
                    error="Failed to append %patchXXXX lines after %setup")

    def get_release_parts(self):
        release = self.get_tag('Release')
        return release_parts(release)

    def recognized_release(self):
        """
        Check if this Release value is something we can parse.
        :rtype: bool
        """
        _, _, rest = self.get_release_parts()
        # If "rest" is not a well-known value here, then this package is
        # using a Release value pattern we cannot recognize.
        if rest == '' or re.match(r'%{\??dist}', rest):
            return True
        return False

    def set_macro(self, macro, value):
        if not RPM_AVAILABLE:
            raise exception.RpmModuleNotAvailable()
        rex = self.RE_MACRO_BASE.format(re.escape(macro))
        rpm.delMacro(macro)
        if value:
            # replace
            self._txt, n = re.subn(r'^(%s).*$' % rex, r'\g<1>%s' % value,
                                   self.txt, flags=re.M)
            if n < 1:
                # create new
                self._txt = u'%global {0} {1}\n{2}'.format(
                    macro, value, self.txt)
            rpm.addMacro(macro, value)
        else:
            # remove
            self._txt = re.sub(r'(^|\n)%s[^\n]+\n?' % rex, r'\g<1>', self.txt)

    def get_macro(self, macro, expanded=False):
        if expanded:
            # XXX: rpm module remembers old values even after .spec change
            # and new Spec() instance (that's why this isn't default)
            return self.expand_macro('%{?' + macro + '}')
        else:
            rex = self.RE_MACRO_BASE.format(re.escape(macro))
            m = re.search('^%s(.*)$' % rex, self.txt, flags=re.M)
            if m:
                v = m.group(1).strip(' \t"')
                return v
            return None

    def set_milestone(self, new_milestone):
        self.set_macro('milestone', new_milestone)

    def get_milestone(self):
        ms = self.get_macro('milestone')
        if ms == '%{?milestone}':
            # counter milestone bug from past rdopkg versions :(
            ms = ''
        return ms

    def set_release(self, new_release, milestone=None, postfix=None):
        release = new_release
        if milestone:
            release += '%{?milestone}'
        self.set_milestone(milestone)
        if postfix is None:
            _, _, postfix = self.get_release_parts()
        release += postfix
        if not re.search(r'%{\??dist}', release):
            release += '%{?dist}'

        return self.set_tag('Release', release)

    def bump_release(self, milestone=None, index=None):
        if index == '0':
            # no bumping
            return
        if not milestone:
            milestone = self.get_milestone()
        numbers, _milestone, postfix = self.get_release_parts()
        if index:
            # case insensitive MAJOR/minor/Patch
            index = index.upper()
        if index is None or index == 'LAST-NUMERIC':
            # bump last numeric only Release part by default
            numlist = numbers.split('.')
            i = -1
            if numbers[-1] == '.':
                i = -2
            numlist[i] = str(int(numlist[i]) + 1)
            release = ".".join(numlist)
        else:
            # bump Nth Release part as specified
            if index in RELEASE_PARTS_SEMVER:
                n = RELEASE_PARTS_SEMVER[index]
            else:
                try:
                    n = int(index)
                except ValueError:
                    raise exception.InvalidReleaseBumpIndex(what=index)
                if n < 0:
                    raise exception.InvalidReleaseBumpIndex(
                        what="%s (positive integer required)" % index)
            # index from 1
            i = n - 1
            release = numbers + _milestone
            parts = release.split('.')
            try:
                parts[i] = str(int(parts[i]) + 1)
            except ValueError:
                raise exception.InvalidReleaseBumpIndex(
                    what="%s. part of Release '%s' isn't numeric: %s" % (
                        n, release, parts[i]))
            except IndexError:
                raise exception.InvalidReleaseBumpIndex(
                    what="%s (Release: %s)" % (
                        n, release))
            release = ".".join(parts)
        return self.set_release(release, milestone=milestone, postfix=postfix)

    def get_vr(self, epoch=None):
        """get VR string from .spec Version, Release and Epoch

        epoch is None: prefix epoch if present (default)
        epoch is True: prefix epoch even if not present (0:)
        epoch is False: omit epoch even if present
        """
        version = self.get_tag('Version', expand_macros=True)
        e = None
        if epoch is None or epoch:
            try:
                e = self.get_tag('Epoch')
            except exception.SpecFileParseError:
                pass
        if epoch is None and e:
            epoch = True
        if epoch:
            if not e:
                e = '0'
            version = '%s:%s' % (e, version)
        release = self.get_tag('Release')
        release = re.sub(r'%\{?\??dist\}?$', '', release)
        release = self.expand_macro(release)
        if release:
            return '%s-%s' % (version, release)
        return version

    def get_nvr(self, epoch=None):
        """get NVR string from .spec Name, Version, Release and Epoch"""
        name = self.get_tag('Name', expand_macros=True)
        vr = self.get_vr(epoch=epoch)
        return '%s-%s' % (name, vr)

    def new_changelog_entry(self, user, email, changes=[]):
        changes_str = "\n".join(map(lambda x: "- %s" % x, changes)) + "\n"
        date = time.strftime('%a %b %d %Y')
        # TODO: detect if there is '-' in changelog entries and use it if so
        vr = self.get_vr()
        head = "* %s %s <%s> %s" % (date, user, email, vr)
        entry = "%s\n%s\n" % (head, changes_str)
        self._txt = re.sub(r'(^%changelog\n)', r'\g<1>%s' % entry,
                           self.txt, count=1, flags=re.M)

    def save(self):
        """ Write the textual content (self._txt) to .spec file (self.fn). """
        if not self.txt:
            # no changes
            return
        if not self.fn:
            raise exception.InvalidAction(
                "Can't save .spec file without its file name specified.")
        f = codecs.open(self.fn, 'w', encoding='utf-8')
        f.write(self.txt)
        f.close()
        self._rpmspec = None

    def get_source_urls(self):
        # arcane rpm constants, now in python!
        sources = list(filter(lambda x: x[2] == 1, self.rpmspec.sources))
        if len(sources) == 0:
            error = "No sources found"
            raise exception.SpecFileParseError(spec_fn=self.fn, error=error)
        # OpenStack packages seem to always use only one tarball
        sources0 = list(filter(lambda x: x[1] == 0, sources))
        if len(sources0) == 0:
            error = "Source0 not found"
            raise exception.SpecFileParseError(spec_fn=self.fn, error=error)
        source_url = sources0[0][0]
        return [source_url]

    def get_source_fns(self):
        return list(map(os.path.basename, self.get_source_urls()))

    def get_last_changelog_entry(self, strip=False):
        changelog = ''
        r = re.split("^%changelog\n", self.txt, flags=re.I | re.M)
        if len(r) > 2:
            raise exception.MultipleChangelog()
        if len(r) == 2:
            changelog = r[1].strip()
        entries = re.split(r'\n\n+', changelog)
        entry = entries[0]
        lines = entry.split("\n")
        if strip:
            lines = list(map(lambda x: x.lstrip(" -*\t"), lines))
        return lines[0], lines[1:]

    def get_requires(self, versions_as_string=False, remove_epoch=True,
                     normalize_py23=False):
        reqs = defaultdict(set)
        for pkg in self.rpmspec.packages:
            pkg_reqs = pkg.header.dsFromHeader('requirename')
            for req in pkg_reqs:
                m = re.match(r'R (\S+)\s+([=<>!]+)\s*(\S+)', req.DNEVR())
                if m:
                    name, eq, ver = m.groups()
                    if eq == '=':
                        eq = '=='
                    if remove_epoch:
                        _, sep, rest = ver.partition(':')
                        if sep:
                            ver = rest
                    if normalize_py23:
                        name = re.sub(r'^python[23]-', 'python-', name)
                    reqs[name].add(eq + ' ' + ver)
                else:
                    name = req.N()
                    if normalize_py23:
                        name = re.sub(r'^python[23]-', 'python-', name)
                    reqs[name]
        if versions_as_string:
            for name in reqs:
                reqs[name] = ','.join(reqs[name])
        return reqs



from rdopkg.action import Action, Arg


ACTIONS = [
    Action('kojibuild', help="build package in koji",
           steps=[
               Action('get_package_env', module='distgit'),
               Action('koji_build'),
           ]),
    Action('cbsbuild', help="build package in CBS",
           steps=[
               Action('get_package_env', module='distgit'),
               Action('cbs_build'),
           ],
           optional_args=[
               Arg('scratch', action='store_true',
                   help='Perform a scratch build'),
           ]),
    Action('mockbuild', help="Run fedpkg/rhpkg mockbuild",
           steps=[
               Action('get_package_env', module='distgit'),
               Action('fedpkg_mockbuild'),
           ]),
]



import os
from rdopkg import helpers
import rdopkg.utils.cmd
import rdopkg.utils.git
from rdopkg.utils.git import git_branch
from rdopkg.utils import specfile


SAMPLE_SPEC = u"""
Name:             {name}
Epoch:            1
Version:          {version}
Release:          {release}
Summary:          Amazing {name} package

Group:            Development/Languages
License:          ASL 2.0
URL:              http://pypi.python.org/pypi/%{{name}}
Source0:          http://pypi.python.org/lul/%{{name}}/%{{name}}-%{{version}}.tar.gz

{magic_comments}

BuildArch:        noarch
BuildRequires:    python-setuptools
BuildRequires:    python2-devel

Requires:         python-argparse
Requires:         python-iso8601
Requires:         python-prettytable

%description
{name} is incredibly interesting and useful to all beings in universe.

It's pretty good stuff for a testing package that doesn't even exist.

%prep
%setup -q

%build
%{{__python}} setup.py build

%install
%{{__python}} setup.py install -O1 --skip-build --root %{{buildroot}}

%files
%doc README.rst
%{{_bindir}}/{name}

%changelog
* Mon Apr 23 2017 Jakub Ruzicka <jruzicka@redhat.com> {version}-{release}
- Update to upstream {version}
- Introduce new bugs (rhbz#123456)

* Tue Mar 23 2016 Jakub Ruzicka <jruzicka@redhat.com> 1.2.2-1
- Update to upstream 1.2.2
- Escape macros in %%changelog
- Counter %%{{?milestone}} bug

* Tue Jun 23 2015 Jakub Ruzicka <jruzicka@redhat.com> 1.1.1-1
- Update to upstream 1.1.1
- Lorem Ipsum
- Dolor sit Amet
"""  # noqa


# silent run() by default
def run(*args, **kwargs):
    if 'log_cmd' not in kwargs:
        kwargs['log_cmd'] = False
    return rdopkg.utils.cmd.run(*args, **kwargs)


class SilentGit(rdopkg.utils.git.Git):
    def __call__(self, *args, **kwargs):
        if kwargs.get('log_cmd', None) is None:
            kwargs['log_cmd'] = False
        return super(SilentGit, self).__call__(*args, **kwargs)


# silent git() by default
git = SilentGit()


def do_patch(fn, content, msg):
    f = open(fn, 'a')
    content = rdopkg.utils.cmd.encode(content)
    f.write(content)
    f.close()
    git('add', fn)
    git('commit', '-m', msg)


def add_patches(patches):
    if not patches:
        return
    for pn in patches:
        fn = pn.replace('\s', '_')
        do_patch(fn, pn + '\n', pn)


def add_n_patches(n, patch_name='Test Patch %d'):
    if not n:
        return
    for i in range(1, n + 1):
        pn = patch_name % i
        fn = pn.lower().replace(' ', '_')
        do_patch(fn, pn + '\n', pn)


def create_sample_distgit(name, version='1.2.3', release='1', path=None,
                          magic_comments=None):
    if not path:
        path = name
    assert not os.path.exists(path)
    if "%{?dist}" not in release:
        release = release + "%{?dist}"
    os.makedirs(path)
    if not magic_comments:
        magic_comments = ''
    with helpers.cdir(path):
        txt = SAMPLE_SPEC.format(name=name, version=version, release=release,
                                 magic_comments=magic_comments)
        spec = specfile.Spec(fn='%s.spec' % name, txt=txt)
        spec.set_tag('Name', name)
        spec.save()
        git('init',)
        git('add', '.')
        git('commit', '-m', 'Initial import', isolated=True)
    return os.path.abspath(path)


def create_sample_patches_branch(n=0, patches=None, version_tag=True):
    version = specfile.Spec().get_tag('Version')
    branch = git.current_branch()
    if version_tag:
        git('tag', version, fatal=False, log_fail=False)
    git('checkout', '-b', '%s-patches' % branch)
    add_patches(patches)
    add_n_patches(n, patch_name="Original Patch %d")
    git('checkout', branch)


def create_sample_upstream_new_version(
        new_version, n_patches, n_from_patches_branch,
        version_tag=True):
    branch = rdopkg.utils.git.git.current_branch()
    old_version = specfile.Spec().get_tag('Version')
    git('checkout', '-b', 'upstream', old_version)
    add_n_patches(n_patches,
                  patch_name='Upstream Commit %d')
    for i in range(n_from_patches_branch):
        # emulate upstream patches that were backported
        git('cherry-pick', 'master-patches' + i * '~')
    if version_tag:
        git('tag', new_version)
    git('checkout', branch)



"""
This is the main rdopkg action module with actions for distgit management.
"""
from __future__ import print_function
from distroinfo.query import get_package
import itertools
import os
import re
import six
from six.moves import input
import sys

from rdopkg.conf import cfg, cfg_files
from rdopkg import exception
from rdopkg import guess
from rdopkg.actionmods import rdoinfo
from rdopkg.actionmods import rpmfactory
from rdopkg.actions.reqs.actions import reqdiff
from rdopkg.utils import log
from rdopkg.utils.cmd import run
from rdopkg.utils.git import git
from rdopkg.utils.issues import search_bug_references
from rdopkg.utils import specfile
from rdopkg.utils import tidy_ssh_user
from rdopkg import helpers

FEDPKG = ['fedpkg']


def get_package_env(version=None, release=None, dist=None, branch=None,
                    patches_branch=None, local_patches_branch=None,
                    patches_style=None, gerrit_patches_chain=None,
                    release_bump_index=None):
    if not branch:
        branch = git.current_branch()
    if branch.endswith('-patches'):
        branch = branch[:-8]
        if git.branch_exists(branch):
            log.info(
                "This looks like -patches branch. Assuming distgit branch: "
                "%s" % branch)
            git.checkout(branch)
        else:
            raise exception.InvalidUsage(
                why="This action must be run on a distgit branch.")
    args = {
        'package': guess.package(),
        'branch': branch,
    }
    osdist = guess.osdist()
    if osdist.startswith('RH'):
        log.info("RH package detected.")
        args['fedpkg'] = ['rhpkg']
    if not patches_branch:
        patches_branch = guess.patches_branch(branch, pkg=args['package'],
                                              osdist=osdist)
    if not patches_style:
        patches_style = guess.patches_style(gerrit_patches_chain)
    args['patches_style'] = patches_style
    args['patches_branch'] = patches_branch
    if release_bump_index is None:
        args['release_bump_index'] = guess.release_bump_index()
    if not local_patches_branch:
        args['local_patches_branch'] = patches_branch.partition('/')[2]
    if not version:
        base_ref = guess.patches_base_ref()
        version, _ = guess.tag2version(base_ref)
        args['version'] = version
    args['version_tag_style'] = guess.version_tag_style(version=version)

    return args


def show_package_env(package, version,
                     branch, patches_branch, local_patches_branch,
                     release_bump_index,
                     version_tag_style=None, patches_style=None,
                     gerrit_patches_chain=None):
    def _putv(title, val):
        print("{t.bold}{title}{t.normal} {val}"
              .format(title=title, val=val, t=log.term))

    osdist = guess.osdist()

    upstream_branch = guess.upstream_branch()
    if not git.ref_exists('refs/remotes/%s' % upstream_branch):
        upstream_version = log.term.yellow('upstream remote/branch not found')
    else:
        upstream_version = guess.upstream_version(branch=upstream_branch)
        if not upstream_version:
            upstream_version = log.term.yellow('no version tag found')

    remote_hash = git.get_latest_commit_hash(patches_branch)
    local_hash = git.get_latest_commit_hash(local_patches_branch)
    upstream_hash = git.get_latest_commit_hash(upstream_branch)
    if remote_hash:
        remote_str = ("{t.green}{hash}{t.normal}"
                      .format(hash=remote_hash[:6], t=log.term))
    else:
        remote_str = log.term.red("not found")
    if local_hash:
        color = 'yellow'
        if local_hash == remote_hash:
            color = 'green'
        local_str = (("{t.%s}{hash}{t.normal}" % color)
                     .format(hash=local_hash[:6], t=log.term))
    else:
        local_str = log.term.yellow("not found")
    if upstream_hash:
        upstream_str = ("{t.green}{hash}{t.normal}"
                        .format(hash=upstream_hash[:6], t=log.term))
    else:
        upstream_str = log.term.yellow("not found")

    if patches_style == 'review':
        if not gerrit_patches_chain:
            gerrit_patches_chain = guess.gerrit_patches_chain(verbose=False)
        gerrit_review_url = rpmfactory.review_url(gerrit_patches_chain) or \
            'unknown'

    spec = specfile.Spec()
    vr = spec.get_vr()
    spec_version = spec.get_tag('Version')
    patches_apply_method = spec.patches_apply_method()
    patches_base, pbn = spec.get_patches_base()
    patches_base_str = patches_base or ''
    if pbn:
        patches_base_str += "{t.magenta}+{n}{t.normal}".format(
            n=pbn, t=log.term)
    if not patches_base_str:
        patches_base_str = "N/A"
    patches_base_tag = guess.version2tag(version, version_tag_style)
    patches_base_ref_str = patches_base_tag
    if git.ref_exists('refs/tags/' + patches_base_tag):
        pbref_exists = '{t.green}existing git tag{t.normal}'
    else:
        ot = git.object_type(patches_base_tag)
        if ot:
            pbref_exists = '{t.green}existing git %s{t.normal}' % ot
        else:
            pbref_exists = '{t.red}invalid git reference{t.normal}'
    patches_base_ref_str += ' : ' + pbref_exists.format(t=log.term)

    release = spec.get_tag('Release')
    release_style = guess.release_style()
    for name, i in specfile.RELEASE_PARTS_SEMVER.items():
        if release_bump_index == str(i):
            # show release part name as well if available (MAJOR/MINOR/PATCH)
            release_bump_index = '%s / %s' % (release_bump_index, name)
            break

    print('')
    _putv('Package:       ', package)
    _putv('VR:            ', vr)
    _putv('Version:       ', spec_version)
    _putv('Upstream:      ', upstream_version)
    _putv('Release:       ', release)
    _putv('Release style: ', release_style)
    _putv('Rls bump index:', release_bump_index)
    _putv('Tag style:     ', version_tag_style or 'X.Y.Z')
    print('')
    _putv('Dist-git branch:       ', branch)
    _putv('Patches style:         ', patches_style)
    _putv('Patches base:          ', patches_base_str)
    _putv('Patches base ref:      ', patches_base_ref_str)
    _putv('Local patches branch:  ',
          '%s : %s' % (local_patches_branch, local_str))
    _putv('Remote patches branch: ',
          '%s : %s' % (patches_branch, remote_str))
    _putv('Remote upstream branch:',
          '%s : %s' % (upstream_branch, upstream_str))
    if patches_style == 'review':
        _putv('Patches chain:         ', gerrit_review_url)
    print('')
    _putv('OS dist:               ', osdist)
    _putv('Patches apply method:  ', patches_apply_method)


def _print_patch_log(patches, tag, n_excluded):
    n_patches = len(patches)
    print("\n{t.bold}{n} patches{t.normal} on top of {t.bold}{tag}{t.normal}"
          ", {t.bold}{ne}{t.normal} excluded\n".format(
              t=log.term, n=n_patches, tag=tag, ne=n_excluded))
    if n_patches <= 0:
        return
    ei = n_patches - n_excluded
    for hsh, title, bzs in patches:
        if ei > 0:
            chsh = log.term.green(hsh)
        else:
            chsh = log.term.red(hsh)
        bzlog = ' '.join(map(lambda x: 'rhbz#%s' % x, bzs))
        patchlog = "%s  %s" % (chsh, title)
        if bzlog != '':
            patchlog += ' (%s)' % bzlog
        print(patchlog)
        ei -= 1


def show_patch_log(version, patches_branch, version_tag_style=None):
    tag = guess.version2tag(version, version_tag_style)
    patches = git.get_commit_bzs(tag, patches_branch)
    spec = specfile.Spec()
    n_excluded = spec.get_n_excluded_patches()
    print("\nPatches branch {t.bold}{pb}{t.normal} is at version {t.bold}"
          "{ver}{t.normal}".format(
              t=log.term, pb=patches_branch, ver=version))
    _print_patch_log(patches, tag, n_excluded)


def conf():
    if cfg_files:
        log.info("Following config files were read:")
        helpers.print_list(cfg_files)
    else:
        log.info("No rdopkg config files found, using default config:")
    log.info("")
    for item in cfg.items():
        log.info("%s: %s" % item)


def new_version_setup(patches_branch=None, local_patches=False,
                      version=None, new_version=None, version_tag_style=None,
                      new_sources=None, no_new_sources=None, unattended=False,
                      bug=None, bump_only=False):
    args = {}
    if new_version:
        # support both version and tag
        ver, _ = guess.tag2version(new_version)
        if ver != new_version:
            new_version = ver
            args['new_version'] = new_version
        new_version_tag = guess.version2tag(new_version, version_tag_style)
        if not bump_only and not git.object_type(new_version_tag):
            raise exception.InvalidGitRef(ref=new_version_tag)
    else:
        ub = guess.upstream_branch()
        if not git.ref_exists('refs/remotes/%s' % ub):
            msg = ("Upstream branch not found: %s\n"
                   "Can't guess latest version.\n\n"
                   "a) provide new version (git tag) yourself\n"
                   "   $ rdopkg new-version 1.2.3\n\n"
                   "b) add upstream git remote:\n"
                   "   $ git remote add -f upstream GIT_URL\n"
                   % ub)
            raise exception.CantGuess(msg=msg)
        new_version_tag = git.get_latest_tag(ub)
        new_version, _ = guess.tag2version(new_version_tag)
        args['new_version'] = new_version
        log.info("Latest version detected from %s: %s" % (ub, new_version))
    if version == new_version:
        if unattended:
            log.info("Package is already at version %s\n" % version)
            raise exception.UserAbort(exit_code=0)
        helpers.confirm(
            msg="It seems the package is already at version %s\n\n"
                "Run new-version anyway?" % version,
            default_yes=False)
    changelog = 'Update to %s' % new_version
    if bug:
        changelog += ' (%s)' % bug
    args['changes'] = [changelog]
    args['new_patches_base'] = new_version_tag
    spec = specfile.Spec()
    rpm_version = spec.get_tag('Version')
    rpm_milestone = spec.get_milestone()
    new_rpm_version, new_milestone = specfile.version_parts(new_version)
    args['new_rpm_version'] = new_rpm_version
    if new_milestone:
        args['new_milestone'] = new_milestone
    if (rpm_version != new_rpm_version
            or bool(new_milestone) != bool(rpm_milestone)):
        if new_milestone:
            args['new_release'] = '0.1'
        else:
            args['new_release'] = '1'
    if not local_patches and not bump_only:
        if not patches_branch or \
           not git.ref_exists('refs/remotes/' + patches_branch):
            log.warn("Patches branch '%s' not found. Running in --bump-only "
                     "mode." % patches_branch)
            args['bump_only'] = True
    if new_sources or no_new_sources:
        if new_sources and no_new_sources:
            raise exception.InvalidUsage(
                msg="DOES NOT COMPUTE: both -n and -N don't make sense.")
        # new_sources == not no_new_sources
    else:
        new_sources = guess.new_sources()
    args['new_sources'] = new_sources

    return args


def ensure_patches_base_ref(bump_only=False):
    if bump_only:
        return
    bref = guess.patches_base_ref()
    if not git.object_type(bref):
        raise exception.InvalidPatchesBaseRef(ref=bref)


def ensure_patches_branch(patches_branch=None, local_patches=False,
                          bump_only=False, patches_style=None):
    if local_patches or bump_only:
        return
    if not patches_branch:
        raise exception.CantGuess(
            what='remote patches branch',
            why="Specify with --patches-branch or use --local-patches")
    if not git.ref_exists('refs/remotes/%s' % patches_branch):
        raise exception.ConfigError(
            what=("remote patches branch '%s' not found. \n\n"
                  "Specify with -p/--patches-branch, use -l/--local-patches, "
                  "or skip patches branch operations with -b/--bump-only" %
                  patches_branch))


def clone(
        package,
        force_fetch=False,
        use_master_distgit=False,
        gerrit_remotes=False,
        review_user=None,
        distro='rdo'):
    rdo = rdoinfo.get_distroinfo(distro=distro)
    ri = rdo.get_info()
    pkg = get_package(ri, package)
    if not pkg:
        raise exception.InvalidPackage(package=package)
    if use_master_distgit:
        try:
            distgit = pkg['master-distgit']
            distgit_str = 'master-distgit'
        except KeyError:
            raise exception.InvalidUsage(
                msg="-m/--use-master-distgit used but 'master-distgit' "
                    "missing in distroinfo for package: %s" % package)
    else:
        distgit = pkg['distgit']
        distgit_str = 'distgit'
    log.info("Cloning {dg} into ./{t.bold}{pkg}{t.normal}/".format(
        t=log.term, dg=distgit_str, pkg=package))
    patches = pkg.get('patches')
    upstream = pkg.get('upstream')
    review_patches = pkg.get('review-patches')
    review_origin = pkg.get('review-origin')

    git('clone', distgit, package)
    with helpers.cdir(package):
        if gerrit_remotes:
            log.info('Adding gerrit-origin remote...')
            git('remote', 'add', 'gerrit-origin', distgit)
        if patches:
            log.info('Adding patches remote...')
            git('remote', 'add', 'patches', patches)
            if gerrit_remotes:
                log.info('Adding gerrit-patches remote...')
                git('remote', 'add', 'gerrit-patches', patches)
        else:
            log.warn("'patches' remote information not available"
                     " in distroinfo.")
        if upstream:
            log.info('Adding upstream remote...')
            git('remote', 'add', 'upstream', upstream)
        else:
            log.warn("'upstream' remote information not available in"
                     " distroinfo.")
        if patches or upstream:
            git('fetch', '--all')

        if not review_user:
            # USERNAME is an env var used by gerrit
            review_user = os.environ.get('USERNAME') or os.environ.get('USER')
        msg_user = ('Using {t.bold}{u}{t.normal} as gerrit username, '
                    'you can change it with '
                    '{t.cmd}git remote set-url {r} ...{t.normal}')
        if review_patches:
            log.info('Adding gerrit remote for patch chains reviews...')
            r = tidy_ssh_user(review_patches, review_user)
            log.info(msg_user.format(u=review_user, r='review-patches',
                                     t=log.term))
            git('remote', 'add', 'review-patches', r)
        else:
            log.warn("'review-patches' remote information not available"
                     " in distroinfo.")
        if review_origin:
            log.info('Adding gerrit remote for reviews...')
            r = tidy_ssh_user(review_origin, review_user)
            log.info(msg_user.format(u=review_user, r='review-origin',
                                     t=log.term))
            git('remote', 'add', 'review-origin', r)
        else:
            log.warn("'review-origin' remote information not available"
                     " in distroinfo.")
        git('remote', '-v', direct=True)


def diff(version, new_version, bump_only=False, no_diff=False,
         version_tag_style=None, unattended=False):
    if bump_only or no_diff or unattended:
        return
    vtag_from = guess.version2tag(version, version_tag_style)
    vtag_to = guess.version2tag(new_version, version_tag_style)
    git('--no-pager', 'diff', '--stat', '%s..%s' % (vtag_from, vtag_to),
        direct=True)
    try:
        reqdiff(vtag_from, vtag_to)
    except Exception:
        pass
    input("Press <Enter> to continue after you inspected the diff. ")


def get_diff_range(diff_range=None, patches_branch=None, branch=None):
    vtag_from, vtag_to = None, None
    if diff_range:
        n = len(diff_range)
        if n > 2:
            raise exception.InvalidUsage(why="diff only supports one or two "
                                             "positional parameters.")
        if n == 2:
            vtag_from, vtag_to = diff_range
        else:
            vtag_to = diff_range[0]
    if not vtag_from:
        if not patches_branch:
            if not branch:
                branch = guess.current_branch()
            patches_branch = guess.patches_branch(branch)
        if not git.ref_exists('refs/remotes/%s' % patches_branch):
            msg = ("Patches branch not found: %s\n"
                   "Can't guess current version.\n\n"
                   "a) provide git tags/refs yourself a la:\n"
                   "   $ rdopkg reqdiff 1.1.1 2.2.2\n\n"
                   "b) add git remote with expected patches branch"
                   % patches_branch)
            raise exception.CantGuess(msg=msg)
        vtag_from = git.get_latest_tag(branch=patches_branch)
    if not vtag_to:
        upstream_branch = guess.upstream_branch()
        vtag_to = git.get_latest_tag(branch=upstream_branch)
    return {
        'version_tag_from': vtag_from,
        'version_tag_to': vtag_to
    }


def _ensure_branch(branch):
    if not branch:
        return
    if git.current_branch() != branch:
        git.checkout(branch)


def _reset_branch(branch, remote_branch):
    if git.branch_exists(branch):
        git('update-ref', 'refs/heads/%s' % branch,
            'refs/remotes/%s' % remote_branch)
    else:
        git.create_branch(branch, remote_branch)


def _is_same_commit(ref1, ref2):
    h1 = git('rev-parse', ref1, log_cmd=False)
    h2 = git('rev-parse', ref2, log_cmd=False)
    return h1 and h1 == h2


def fetch_all():
    git('fetch', '--all', direct=True)


def prep_new_patches_branch(new_version,
                            local_patches_branch, patches_branch,
                            local_patches=False, bump_only=False,
                            patches_style=None, version_tag_style=None,
                            unattended=False, no_push_patches=False):
    if patches_style == 'review':
        if no_push_patches:
            return
        new_version_tag = guess.version2tag(new_version, version_tag_style)
        try:
            remote, branch = git.remote_branch_split(patches_branch)
            if unattended:
                log.warn('Unattended mode: force pushing patches')
            else:
                helpers.confirm("Push %s to %s/%s (with --force)?" % (
                    new_version_tag, remote, branch))
            git('branch', '--force', local_patches_branch, new_version_tag)
            git('push', '--force', remote,
                '%s:%s' % (local_patches_branch, branch))
            # push the tag
            git('push', '--force', remote, new_version_tag)
        except exception.UserAbort:
            pass
    else:
        if not (local_patches or bump_only):
            _reset_branch(local_patches_branch, remote_branch=patches_branch)


def get_patches_branch(local_patches_branch, patches_branch,
                       local_patches=False, patches_style=None,
                       gerrit_patches_chain=None,
                       bump_only=False, force=False):
    if local_patches or bump_only:
        return
    if patches_style == 'review':
        if not gerrit_patches_chain:
            gerrit_patches_chain = guess.gerrit_patches_chain()
        if gerrit_patches_chain:
            rpmfactory.fetch_patches_branch(local_patches_branch,
                                            gerrit_patches_chain,
                                            force)
        else:
            log.warn("Review patches chain not found. No patches yet?")
    else:
        _reset_branch(local_patches_branch, remote_branch=patches_branch)


def checkout_patches_branch(local_patches_branch):
    git.checkout(local_patches_branch)


def review_patches_branch(local_patches_branch, patches_style=None,
                          bump_only=False, unattended=False):
    if patches_style != 'review' or bump_only:
        return
    try:
        if unattended:
            log.warn("Unattended mode: sending %s branch for review")
        else:
            helpers.confirm("Send %s branch for review?" %
                            local_patches_branch)
        rpmfactory.review_patch(local_patches_branch)
    except exception.UserAbort:
        pass


def rebase_patches_branch(new_version, local_patches_branch,
                          patches_branch=None, local_patches=False,
                          patches_style=None, version_tag_style=None,
                          bump_only=False):
    if bump_only:
        return
    git.checkout(local_patches_branch)
    new_version_tag = guess.version2tag(new_version, version_tag_style)
    git('rebase', new_version_tag, direct=True)

    if patches_style != 'review':
        if local_patches or not patches_branch:
            return
        if _is_same_commit(local_patches_branch, patches_branch):
            log.info("%s is up to date, no need for push." % patches_branch)
            return
        try:
            remote, branch = git.remote_branch_split(patches_branch)
            helpers.confirm("Push %s to %s / %s (with --force)?" % (
                local_patches_branch, remote, branch))
            git('push', '--force', remote,
                '%s:%s' % (local_patches_branch, branch))
            # push the tag
            git('push', '--force', remote, new_version_tag)
        except exception.UserAbort:
            pass


def check_new_patches(version, local_patches_branch,
                      patches_style=None, local_patches=False,
                      patches_branch=None, changes=None,
                      version_tag_style=None, changelog=None,
                      no_bump=False):
    if no_bump:
        return
    if not changes:
        changes = []
    if changelog:
        changelog = changelog.lower()
    else:
        # default changelog format
        changelog = 'detect'

    if changelog == 'detect' or changelog == 'count':
        if local_patches or patches_style == 'review':
            head = local_patches_branch
        else:
            if not patches_branch:
                raise exception.RequiredActionArgumentNotAvailable(
                    action='check_new_patches', arg='patches_branch')
            head = patches_branch

        version_tag = guess.version2tag(version, version_tag_style)
        patches = git.get_commit_bzs(version_tag, head)
        old_patches = specfile.get_patches_from_files()
        spec = specfile.Spec()

        n_git_patches = len(patches)
        n_ignore_patches = 0
        ignore_regex = spec.get_patches_ignore_regex()
        if ignore_regex:
            patches = (flatten(_partition_patches(patches, ignore_regex)))
            n_ignore_patches = n_git_patches - len(patches)

        n_skip_patches = spec.get_n_excluded_patches()
        if n_skip_patches > 0:
            patches = patches[0:-n_skip_patches]

        log.debug("Total patches in git:%d skip:%d ignore:%d" % (
            n_git_patches, n_skip_patches, n_ignore_patches))

        if changelog == 'detect':
            # detect new/old patches by hash/subject
            def _patch_filter(c):
                hash, subj, bzs = c
                for _, old_hash, old_subj in old_patches:
                    if helpers.is_same_hash(hash, old_hash):
                        return False
                    if helpers.is_same_subject(subj, old_subj):
                        # maybe more relaxed match on subjects?
                        return False
                return True
            patches = filter(_patch_filter, patches)

        elif changelog == 'count':
            # assume no removed patches, include new ones in changelog
            n_base_patches = n_skip_patches + spec.get_n_patches()
            if n_base_patches > 0:
                patches = patches[0:-n_base_patches]

        for _, subj, bzs in patches:
            bzstr = ' '.join(map(lambda x: 'rhbz#%s' % x, bzs))
            if bzstr:
                subj += ' (%s)' % bzstr
            changes.append(subj)

    elif changelog == 'plain':
        # just leave a generic message without inspecting commits (see below)
        pass
    else:
        raise exception.InvalidUsage(
            why="Not a valid changelog format: %s" % changelog)

    if not changes:
        changes.append('Update patches')
    return {'changes': changes}


def get_upstream_patches(version, local_patches_branch,
                         patches_branch=None, upstream_branch=None,
                         new_milestone=None):
    # TODO: nuke this, looks unused
    patches = git("log", "--cherry-pick", "--pretty=format:\%s",
                  "%(remote)s...%(local)s" % {'remote': patches_branch,
                                              'local': local_patches_branch})
    changes = [p.strip().replace('\\', '')
               for p in patches.split('\n') if p != '']

    if not changes:
        log.warn("No new patches detected in %s." % local_patches_branch)
        helpers.confirm("Do you want to continue anyway?", default_yes=False)

    n_patches = len(changes)
    changes.insert(0, ("Rebase %s changes from %s" %
                       (n_patches, upstream_branch)))
    args = {'changes': changes}
    if n_patches > 0:
        if new_milestone:
            new_milestone += '.p%d' % n_patches
        else:
            new_milestone = 'p%d' % n_patches
        args['new_milestone'] = new_milestone
    return args


def update_spec(branch=None, changes=None,
                new_rpm_version=None, new_release=None,
                new_milestone=None, new_patches_base=None,
                no_bump=False, changelog_user=None,
                changelog_email=None, release_bump_index=None):

    if no_bump:
        return
    if not changes:
        changes = []
    _ensure_branch(branch)
    spec = specfile.Spec()
    if new_rpm_version:
        old_version = spec.get_tag('Version')
        if specfile.has_macros(old_version):
            log.info('Version contains macro - not touching that.')
        else:
            spec.set_tag('Version', new_rpm_version)
    if new_release is not None:
        if spec.recognized_release():
            spec.set_release(new_release, milestone=new_milestone)
        else:
            log.info('Custom Release format detected '
                     '- assuming custom milestone management.')
            spec.set_release(new_release)
    else:
        spec.bump_release(milestone=new_milestone,
                          index=release_bump_index)
    if new_patches_base:
        new_patches_base_version, _ = guess.tag2version(new_patches_base)
        if new_patches_base_version == new_rpm_version:
            new_patches_base = None
        changed = spec.set_patches_base_version(new_patches_base)
        if not changed:
            log.info("Macro detected in patches_base - not touching that.")
    user = (changelog_user or guess.user())
    email = (changelog_email or guess.email())
    spec.new_changelog_entry(user=user, email=email,
                             changes=changes)
    spec.save()


def get_source(new_sources=False):
    if not new_sources:
        # due to new-version changes, this function does nothing by default :)
        # TODO: factor this into get_source() and new_version_get_source()
        return
    # TODO: consider using `spectool --gf --source`
    source_urls = specfile.Spec().get_source_urls()
    # So far, only Source/Source0 is a tarball to download
    source_url = source_urls[0]
    source_fn = os.path.basename(source_url)
    if os.path.isfile(source_fn):
        log.info("%s already present" % source_fn)
        return
    try:
        helpers.download_file(source_url)
    except exception.CommandFailed:
        # only valid for new-sources, not for get-source
        raise exception.ActionRequired(
            msg="Failed to download source tarball. Please update Source0 in "
                ".spec file.", rerun=True)


def new_sources(branch=None, fedpkg=FEDPKG, new_sources=False):
    _ensure_branch(branch)
    if not new_sources:
        return
    sources = specfile.Spec().get_source_fns()
    cmd = fedpkg + ['new-sources'] + sources
    run(*cmd, direct=True)


def _commit_message(changes=None, header_file=None, no_bump=False,
                    local_patches_branch=None, amend=False):
    msg = None
    if header_file:
        # use supplied header file
        try:
            if header_file == '-':
                msg = sys.stdin.read()
            else:
                msg = open(header_file).read()
        except IOError as ex:
            raise exception.FileNotFound(msg=str(ex))
    else:
        if no_bump:
            # generic message for changes that don't generate changelog
            if local_patches_branch:
                msg = 'Updated patches from %s' % local_patches_branch
            else:
                msg = 'Updated patches'
        else:
            # use NVR as default commit message header
            msg = specfile.Spec().get_nvr(epoch=False)

    if no_bump:
        # no new changelog entry for this change
        return msg
    # append changelog generated from changes argument
    if not changes:
        # by default, use last .spec %changelog entry lines
        _, changes = specfile.Spec().get_last_changelog_entry(strip=True)
        if not changes:
            raise exception.IncompleteChangelog()
    changes_str = "\n".join(map(lambda x: "- %s" % x, changes))
    msg += "\n\nChangelog:\n%s" % changes_str
    # append Resloves: rhbz#12345 lines mentioned in changes
    fixed_rhbzs = []
    for change in changes:
        for n in search_bug_references(change):
            if n not in fixed_rhbzs:
                fixed_rhbzs.append(n)
    if fixed_rhbzs:
        rhbzs_str = "\n".join(map(lambda x: "Resolves: rhbz#%s" % x,
                                  fixed_rhbzs))
        msg += "\n\n%s" % rhbzs_str
    if amend:
        amendables = ['Change-Id', ]
        return _add_amendables(msg, amendables)
    return msg


def _add_amendables(msg, amendables):
    # Rather than go with an OrderedDict, keep it simple with a list
    AMENDABLES_ORDER = ['Change-Id', ]
    changeid_regex = 'Change-Id: (?P<id>I[a-fA-F0-9]+)'
    AMENDABLES = {'Change-Id': {'regex': re.compile(changeid_regex),
                                'msg': '%(msg)s\n\nChange-Id: %(id)s\n'}, }

    previous_commit = git.current_commit_message()
    for unknown in (set(amendables).difference(AMENDABLES_ORDER)):
        log.warn('"%s" is not a valid amendable field' % unknown)
    for a in AMENDABLES_ORDER:
        if a not in amendables:
            continue
        regex = AMENDABLES[a]['regex']
        if regex.search(previous_commit):
            rdict = regex.search(previous_commit).groupdict()
            rdict['msg'] = msg
            msg = AMENDABLES[a]['msg'] % rdict
    return msg


def commit_distgit_update(branch=None, local_patches_branch=None,
                          amend=False, no_bump=False, commit_header_file=None):
    _ensure_branch(branch)
    if git.is_clean():
        raise exception.NoDistgitChangesFound()
    msg = _commit_message(header_file=commit_header_file,
                          no_bump=no_bump,
                          local_patches_branch=local_patches_branch,
                          amend=amend)
    cmd = ['commit', '-a', '-F', '-']
    if amend:
        cmd.append('--amend')
    git(*cmd, input=msg, print_output=True)


def amend(commit_header_file=None):
    msg = _commit_message(header_file=commit_header_file, amend=True)
    git('commit', '-a', '--amend', '-F', '-', input=msg, print_output=True)
    print("")
    git('--no-pager', 'log', '--name-status', 'HEAD~..HEAD', direct=True)


def _partition_patches(patches, regex):
    if regex is None:
        return [patches]

    def take(_patch):
        return not bool(regex.search(_patch[1]))

    def _stacker(buckets):
        while True:
            item, new_bucket = yield
            if new_bucket:
                buckets.append([item])
            else:
                buckets[-1].append(item)

    def _filter(check, stacker):
        start_bucket = True
        while True:
            item = yield
            if check(item):
                stacker.send((item, start_bucket))
                start_bucket = False
            else:
                start_bucket = True

    buckets = []
    stacker = _stacker(buckets)
    six.next(stacker)
    filter = _filter(take, stacker)
    six.next(filter)

    for patch in patches:
        filter.send(patch)
    return buckets


def flatten(list_of_lists):
    return list(itertools.chain(*list_of_lists))


def update_patches(branch, local_patches_branch, bump_only=False,
                   version=None, new_version=None, version_tag_style=None):
    if bump_only:
        return
    target_version = new_version or version
    if not target_version:
        raise exception.RequiredActionArgumentNotAvailable(
            action='update_patches',
            arg='version or new_version')
    tag = guess.version2tag(target_version, version_tag_style)

    _ensure_branch(local_patches_branch)
    patches = list(git.get_commits(tag, local_patches_branch))
    n_patches = len(patches)
    _ensure_branch(branch)
    spec = specfile.Spec()
    spec.sanity_check()
    patches_base, n_excluded = spec.get_patches_base()
    ignore_regex = spec.get_patches_ignore_regex()
    if ignore_regex and patches_base is None:
        # TODO: patches_base and patches_ignore should be independent
        # patches_ignore feature tests should help with this a lot
        raise exception.OnlyPatchesIgnoreUsed()
        pass

    patch_fns = spec.get_patch_fns()
    for pfn in patch_fns:
        git('rm', '-f', '--ignore-unmatch', pfn)
    patch_fns = []

    if n_excluded > 0:
        patches = patches[:-n_excluded]
    patches.reverse()

    ranges = [patches]
    filtered_patches = patches
    if ignore_regex:
        ranges = _partition_patches(patches, ignore_regex)
        filtered_patches = flatten(ranges)
    n_filtered_out = len(patches) - len(filtered_patches)

    if ignore_regex:
        fmt = ('\nUsing {t.bold}patches_ignore={t.normal}{t.magenta}%s'
               '{t.normal} regexp to filter out patches.'
               ) % ignore_regex.pattern
    else:
        fmt = ('\nNo valid {t.bold}patches_ignore{t.normal} '
               'filtering regex found in the .spec file.')
    log.info(fmt.format(t=log.term))
    log.info(
        "\n{t.bold}{n} patches{t.normal} on top of {t.bold}{tag}{t.normal}"
        ", {t.bold}{ne}{t.normal} excluded by base"
        ", {t.bold}{nf}{t.normal} filtered out by regex.".format(
            t=log.term,
            n=n_patches,
            tag=tag,
            ne=n_excluded,
            nf=n_filtered_out))

    if patches and filtered_patches:
        for hsh, title in reversed(filtered_patches):
            log.info("%s  %s" % (log.term.green(hsh), title))

        log.info("")
        patch_fns = []
        for patch_range in ranges:
            start_commit, _title = patch_range[0]
            end_commit, _title = patch_range[-1]
            start_number = len(patch_fns) + 1

            rng = git.rev_range(start_commit + '~', end_commit)
            format_patch_cmd = ['-c', 'core.abbrev=7',
                                'format-patch', '--no-renames',
                                '--no-signature', '-N', '--ignore-submodules',
                                '--stat=80', '--summary',
                                '--start-number', str(start_number), rng]

            o = git(*format_patch_cmd)
            range_files = git._parse_output(o)
            patch_fns.extend(range_files)

        for pfn in patch_fns:
            git('add', pfn)

    spec.set_new_patches(patch_fns)
    patches_branch_ref = git('rev-parse', local_patches_branch, log_cmd=False)
    spec.set_commit_ref_macro(patches_branch_ref)
    spec.save()


def ensure_patches_changed():
    if git.is_clean():
        raise exception.NoDistgitChangesFound(
            msg="No changes to patches found",
            # `rdopkg patch` is a success even on no new patches
            exit_code=0)


def squash():
    git.squash_last()


def final_spec_diff(branch=None):
    _ensure_branch(branch)
    print("Important distgit changes:")
    spec = specfile.Spec()
    git('--no-pager', 'diff', 'HEAD~..HEAD', '--', spec.fn,
        direct=True, fatal=False)
    print("")
    git('--no-pager', 'log', '--name-status', 'HEAD~..HEAD',
        direct=True, fatal=False)
    print("\nRequested distgit update finished, see last commit.")


def edit_spec():
    raise exception.ActionRequired(
        msg="Edit .spec file as needed and describe changes in changelog.")


def make_srpm(package, dist=None, fedpkg=FEDPKG):
    cmd = list(fedpkg)
    if dist:
        dname, _, drls = dist.partition('-')
        if dname == 'epel' and drls:
            cmd += ['--dist', 'el' + drls]
    cmd.append('srpm')
    out = run(*cmd)
    m = re.search(r'/([^/\\]+\.src.rpm)\b', out)
    if not m:
        raise exception.CommandOutputParseError(tool=cmd[0], output=out)
    srpm = m.group(1)
    if not os.path.isfile(srpm):
        raise exception.FileNotFound(path=srpm)
    return {'srpm': srpm}


def tag_patches_branch(package, local_patches_branch, patches_branch,
                       force=False, push=False):
    """ Tag the local_patches_branch with this package's NVR. """
    vr = specfile.Spec().get_vr(epoch=False)
    nvr_tag = package + '-' + vr
    tag_cmd = ['tag', nvr_tag, local_patches_branch]
    if force:
        tag_cmd.append('-f')
    git(*tag_cmd)
    if push:
        patches_remote = patches_branch.partition('/')[0]
        git('push', patches_remote, nvr_tag)
    else:
        print('Not pushing tag. Run "git push patches %s" by hand.' % nvr_tag)



from __future__ import print_function
import contextlib
import os
import re
import six
from six.moves import input
from six import string_types
import sys
import yaml

from rdopkg import exception
from rdopkg.utils import log
from rdopkg.utils.cmd import run

if six.PY2:
    from collections import Iterable
else:
    from collections.abc import Iterable


def confirm(msg, default_yes=True):
    if sys.stdout.isatty():
        if default_yes:
            options = '[Yn]'
        else:
            options = '[Ny]'
        print('')
        i = input(log.term.important("%s %s " % (msg, options))).lower()
        if not i:
            if default_yes:
                return
            else:
                raise exception.UserAbort()
        if i != 'y' and i != 'yes':
            raise exception.UserAbort()
        print('')
    else:
        if default_yes:
            return
        else:
            raise exception.UserAbort()


def download_file(url):
    run('curl', '-L', '-f', '-O', url, direct=True)


def edit(path):
    editor = os.environ.get('EDITOR')
    if not editor:
        editor = 'vim'
        log.info("$EDITOR not set. Falling back to %s." % editor)
    try:
        r = run(editor, path, direct=True)
    except exception.CommandNotFound:
        raise exception.CommandNotFound(
            msg='Failed to find suitable text editor ({0}).  '
                'Please set $EDITOR environment variable.'.format(editor))
    return r.success


def ensure_dir(path):
    if os.path.exists(path):
        if not os.path.isdir(path):
            raise exception.NotADirectory(path=path)
    else:
        os.makedirs(path)


def ensure_new_file_dirs(path):
    if path[-1] == os.sep:
        raise exception.NotAFile(path=path)
    _dir = os.path.dirname(path)
    if _dir:
        ensure_dir(_dir)


def find_files(path, ext=None):
    paths = [os.path.join(path, f) for f in os.listdir(path)
             if os.path.isfile(os.path.join(path, f))]
    if ext:
        paths = [p for p in paths if p.endswith(ext)]
    return paths


def print_list(l, pre='- ', nl_before=False, nl_after=False):
    if nl_before:
        print("")
    for e in l:
        print('%s%s' % (pre, e))
    if nl_after:
        print("")


def action_required(msg):
    print("\n{t.important}Action required: {t.normal}{t.bold}{msg}{t.normal}"
          .format(t=log.term, msg=msg))
    print("\nOnce done, run `rdopkg -c` to continue.\n")


def is_same_hash(h1, h2):
    if h1 is None or h2 is None:
        return False
    if h1 == h2 or h1.startswith(h2) or h2.startswith(h1):
        return True
    return False


def strip_patch_subject(subj):
    return re.sub(r'^\[PATCH\]\s+', '', subj.strip())


def is_same_subject(s1, s2):
    if s1 is None or s2 is None:
        return False
    if strip_patch_subject(s1) == strip_patch_subject(s2):
        return True
    return False


@contextlib.contextmanager
def cdir(path):
    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


@contextlib.contextmanager
def setenv(**env_dict):
    orig_env = os.environ.copy()
    for k, v in env_dict.items():
        os.environ[k] = str(v)
    try:
        yield
    finally:
        for k, v in env_dict.items():
            del os.environ[k]
        for k, v in orig_env.items():
            os.environ[k] = v


def print_keyval(key, val, kb=True, vb=False):
    if kb:
        fmt = '{t.bold}{key}{t.normal}: '
    else:
        fmt = '{key}'
    if vb:
        fmt += '{t.bold}{val}{t.bold}'
    else:
        fmt += '{val}'
    if not isinstance(
            val,
            string_types) and isinstance(
            val,
            Iterable):
        vals = "\n" + "\n".join(map(lambda x: '- ' + str(x), val))
    else:
        vals = str(val)
    print(fmt.format(t=log.term, key=key, val=vals))


class DictPrinter(object):
    # Very simple and stupid dict printer.

    def __init__(self, header=None, first=None, last=None):
        self.first = first or []
        self.last = last or []
        self.header = header

    def __call__(self, d):
        dd = d.copy()
        # print header
        if self.header:
            hdr = dd.pop(self.header, '')
            print_keyval(self.header, hdr, kb=True, vb=True)
        # print first fields
        for key in self.first:
            if key not in dd:
                continue
            val = dd.pop(key)
            print_keyval(key, val)
        # filter out last fields
        last_items = []
        for key in self.last:
            if key not in dd:
                continue
            last_items.append((key, dd.pop(key)))
        if dd:
            print(yaml.dump(dd, default_flow_style=False).rstrip())
        for key, val in last_items:
            print_keyval(key, val)



from rdopkg.utils.issues import search_bug_references
import pytest


def test_patch_parsing_for_rhbz_numbers():
    data = [
        ('Resolves: rhbz#1433685', 'rhbz#1433685', 1433685),
        ('Related: rhbz#1433685', 'rhbz#1433685', 1433685),
        ('  Related:   \t  rhbz#1433685  ', 'rhbz#1433685', 1433685),
    ]

    # TODO logic here is convoluted as bz does not convey enough information
    # here search_bug_references should probably make it clear it's only
    # looking for rhbz#'s or return qualified rhbz#1234 entries and places
    # that consume it handle that cleanly

    for text, _, expected in data:
        assert str(expected) in search_bug_references(text)


def test_patch_parsing_no_input():
    assert not search_bug_references('')


def test_patch_parsing_input_None():
    assert not search_bug_references(None)


def test_patch_parsing_no_bug_data_to_be_found():
    assert not search_bug_references(' foo bar blah bug found 1324')


@pytest.mark.skip(
    'unhandled content seen in wild that could be parsed for bug numbers')
def test_patch_parsing_for_rhbz_numbers_bad_seen_in_wild():
    data_variants_seen_in_wild = [
        ('Related rhbz: 1433685', 'rhbz#1433685', 1433685),
        ('Resolves rhbz: 1433685', 'rhbz#1433685', 1433685),
        ('Resolves: rhbz 1433685', 'rhbz#1433685', 1433685),
        ('Resolves: rhbz# 1433685', 'rhbz#1433685', 1433685),
        ('Resolves: rhbz #1433685', 'rhbz#1433685', 1433685),
    ]

    for text, _, expected in data_variants_seen_in_wild:
        assert str(expected) in search_bug_references(text)
